@@ -8,12 +8,12 @@
 //  SS  SS   tt   ee      ll   ll  aa  aa
 //   SSSS     ttt  eeeee llll llll  aaaaa
 //
-// Copyright (c) 1995-2005 by Bradford W. Mott and the Stella team
+// Copyright (c) 1995-2007 by Bradford W. Mott and the Stella team
 //
 // See the file "license" for information on usage and redistribution of
 // this file, and for a DISCLAIMER OF ALL WARRANTIES.
 //
-// $Id: FrameBufferGL.cxx,v 1.55 2006/01/16 01:23:55 stephena Exp $
+// $Id: FrameBufferGL.cxx,v 1.86 2007/01/15 00:07:51 stephena Exp $
 //============================================================================
 
 #ifdef DISPLAY_OPENGL
@@ -31,6 +31,22 @@
 #include "Font.hxx"
 #include "GuiUtils.hxx"
 
+// There's probably a cleaner way of doing this
+// These values come from SDL_video.c
+// If they change, this code will break horribly
+#ifndef SDL_GL_ACCELERATED_VISUAL
+  #define SDL_GL_ACCELERATED_VISUAL SDL_GLattr(15)
+#endif
+#ifndef SDL_GL_SWAP_CONTROL
+  #define SDL_GL_SWAP_CONTROL SDL_GLattr(16)
+#endif
+
+// These are not defined in the current version of SDL_opengl.h
+// Hopefully speed up OpenGL rendering in OSX
+#ifndef GL_TEXTURE_RECTANGLE_ARB
+  #define GL_TEXTURE_RECTANGLE_ARB 0x84F5
+#endif
+
 // Maybe this code could be cleaner ...
 static void (APIENTRY* p_glClear)( GLbitfield );
 static void (APIENTRY* p_glEnable)( GLenum );
@@ -38,6 +54,7 @@ static void (APIENTRY* p_glDisable)( GLenum );
 static void (APIENTRY* p_glPushAttrib)( GLbitfield );
 static const GLubyte* (APIENTRY* p_glGetString)( GLenum );
 static void (APIENTRY* p_glHint)( GLenum, GLenum );
+static void (APIENTRY* p_glShadeModel)( GLenum );
 
 // Matrix
 static void (APIENTRY* p_glMatrixMode)( GLenum );
@@ -70,11 +87,11 @@ static void (APIENTRY* p_glTexParameteri)( GLenum, GLenum, GLint );
 FrameBufferGL::FrameBufferGL(OSystem* osystem)
   : FrameBuffer(osystem),
     myTexture(NULL),
+    myHaveTexRectEXT(false),
     myScreenmode(0),
     myScreenmodeCount(0),
-    myTextureID(0),
-    myFilterParam(GL_NEAREST),
     myFilterParamName("GL_NEAREST"),
+    myZoomLevel(1),
     myFSScaleFactor(1.0),
     myDirtyFlag(true)
 {
@@ -86,7 +103,7 @@ FrameBufferGL::~FrameBufferGL()
   if(myTexture)
     SDL_FreeSurface(myTexture);
 
-  p_glDeleteTextures(1, &myTextureID);
+  p_glDeleteTextures(1, &myBuffer.texture);
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -98,7 +115,12 @@ bool FrameBufferGL::loadFuncs(const string& library)
   if(SDL_WasInit(SDL_INIT_VIDEO) == 0)
     SDL_Init(SDL_INIT_VIDEO);
 
-  if(SDL_GL_LoadLibrary(library.c_str()) < 0)
+  // Try both the specified library and auto-detection
+  bool libLoaded = (library != "" && SDL_GL_LoadLibrary(library.c_str()) >= 0);
+  bool autoLoaded = false;
+  if(!libLoaded) autoLoaded = (SDL_GL_LoadLibrary(0) >= 0);
+
+  if(!libLoaded && !autoLoaded)
     return false;
 
   // Otherwise, fill the function pointers for GL functions
@@ -117,6 +139,8 @@ bool FrameBufferGL::loadFuncs(const string& library)
       SDL_GL_GetProcAddress("glGetString"); if(!p_glGetString) return false;
   p_glHint = (void(APIENTRY*)(GLenum, GLenum))
       SDL_GL_GetProcAddress("glHint"); if(!p_glHint) return false;
+  p_glShadeModel = (void(APIENTRY*)(GLenum))
+      SDL_GL_GetProcAddress("glShadeModel"); if(!p_glShadeModel) return false;
 
   p_glMatrixMode = (void(APIENTRY*)(GLenum))
       SDL_GL_GetProcAddress("glMatrixMode"); if(!p_glMatrixMode) return false;
@@ -170,42 +194,16 @@ bool FrameBufferGL::initSubsystem()
   myDepth = SDL_GetVideoInfo()->vfmt->BitsPerPixel;
   switch(myDepth)
   {
-    case 8:
-      myRGB[0] = 3;
-      myRGB[1] = 3;
-      myRGB[2] = 2;
-      myRGB[3] = 0;
-      break;
-
     case 15:
-      myRGB[0] = 5;
-      myRGB[1] = 5;
-      myRGB[2] = 5;
-      myRGB[3] = 0;
-      break;
-
     case 16:
-      myRGB[0] = 5;
-      myRGB[1] = 6;
-      myRGB[2] = 5;
-      myRGB[3] = 0;
+      myRGB[0] = 5; myRGB[1] = 5; myRGB[2] = 5; myRGB[3] = 0;
       break;
-
     case 24:
-      myRGB[0] = 8;
-      myRGB[1] = 8;
-      myRGB[2] = 8;
-      myRGB[3] = 0;
-      break;
-
     case 32:
-      myRGB[0] = 8;
-      myRGB[1] = 8;
-      myRGB[2] = 8;
-      myRGB[3] = 8;
+      myRGB[0] = 8; myRGB[1] = 8; myRGB[2] = 8; myRGB[3] = 0;
       break;
-
     default:  // This should never happen
+      return false;
       break;
   }
 
@@ -226,31 +224,29 @@ bool FrameBufferGL::initSubsystem()
   SDL_GL_GetAttribute( SDL_GL_BLUE_SIZE, (int*)&myRGB[2] );
   SDL_GL_GetAttribute( SDL_GL_ALPHA_SIZE, (int*)&myRGB[3] );
 
-#ifndef TEXTURES_ARE_LOST
-  // Create the texture surface
-  createTextures();
-#endif
-
-  // Show some OpenGL info
-  if(myOSystem->settings().getBool("showinfo"))
-  {
-    cout << "Video rendering: OpenGL mode" << endl;
-
-    ostringstream colormode;
-    colormode << "  Color   : " << myDepth << " bit, " << myRGB[0] << "-"
-              << myRGB[1] << "-" << myRGB[2] << "-" << myRGB[3];
-
-    cout << "  Vendor  : " << p_glGetString(GL_VENDOR) << endl
-         << "  Renderer: " << p_glGetString(GL_RENDERER) << endl
-         << "  Version : " << p_glGetString(GL_VERSION) << endl
-         << colormode.str() << endl
-         << "  Filter  : " << myFilterParamName << endl
-         << endl;
-  }
-
   return true;
 }
 
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+string FrameBufferGL::about()
+{
+  string extensions;
+  if(myHaveTexRectEXT) extensions += "GL_TEXTURE_RECTANGLE_ARB  ";
+  if(extensions == "") extensions = "None";
+
+  ostringstream out;
+  out << "Video rendering: OpenGL mode" << endl
+      << "  Vendor:     " << p_glGetString(GL_VENDOR) << endl
+      << "  Renderer:   " << p_glGetString(GL_RENDERER) << endl
+      << "  Version:    " << p_glGetString(GL_VERSION) << endl
+      << "  Color:      " << myDepth << " bit, " << myRGB[0] << "-"
+      << myRGB[1] << "-"  << myRGB[2] << "-" << myRGB[3] << endl
+      << "  Filter:     " << myFilterParamName << endl
+      << "  Extensions: " << extensions << endl;
+
+  return out.str();
+}
+
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void FrameBufferGL::setAspectRatio()
 {
@@ -259,6 +255,12 @@ void FrameBufferGL::setAspectRatio()
     theAspectRatio = 1.0;
 }
 
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+void FrameBufferGL::setScaler(Scaler scaler)
+{
+  myZoomLevel = scaler.zoom;
+}
+
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 bool FrameBufferGL::createScreen()
 {
@@ -267,12 +269,19 @@ bool FrameBufferGL::createScreen()
   SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE,  myRGB[2] );
   SDL_GL_SetAttribute( SDL_GL_ALPHA_SIZE, myRGB[3] );
   SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
+  SDL_GL_SetAttribute( SDL_GL_ACCELERATED_VISUAL, 1 );
+
+  // There's no guarantee this is supported on all hardware
+  // We leave it to the user to test and decide
+  int vsync = myOSystem->settings().getBool("gl_vsync") ? 1 : 0;
+  SDL_GL_SetAttribute( SDL_GL_SWAP_CONTROL, vsync );
 
   // Set the screen coordinates
   GLdouble orthoWidth  = 0.0;
   GLdouble orthoHeight = 0.0;
   setDimensions(&orthoWidth, &orthoHeight);
 
+  // Create screen containing GL context
   myScreen = SDL_SetVideoMode(myScreenDim.w, myScreenDim.h, 0, mySDLFlags);
   if(myScreen == NULL)
   {
@@ -280,33 +289,35 @@ bool FrameBufferGL::createScreen()
     return false;
   }
 
-  p_glPushAttrib(GL_ENABLE_BIT);
+  // Check for some extensions that can potentially speed up operation
+  const char* extensions = (const char *) p_glGetString(GL_EXTENSIONS);
+  myHaveTexRectEXT       = strstr(extensions, "ARB_texture_rectangle") != NULL;
 
-  // Center the image horizontally and vertically
+  // Initialize GL display
   p_glViewport(myImageDim.x, myImageDim.y, myImageDim.w, myImageDim.h);
+  p_glShadeModel(GL_FLAT);
+  p_glDisable(GL_CULL_FACE);
+  p_glDisable(GL_DEPTH_TEST);
+  p_glDisable(GL_ALPHA_TEST);
+  p_glDisable(GL_LIGHTING);
+  p_glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
 
   p_glMatrixMode(GL_PROJECTION);
-  p_glPushMatrix();
   p_glLoadIdentity();
-
-  p_glOrtho(0.0, orthoWidth, orthoHeight, 0.0, 0.0, 1.0);
-
+  p_glOrtho(0.0, orthoWidth, orthoHeight, 0, -1.0, 1.0);
   p_glMatrixMode(GL_MODELVIEW);
-  p_glPushMatrix();
   p_glLoadIdentity();
 
-#ifdef TEXTURES_ARE_LOST
+  // Allocate GL textures
   createTextures();
-#endif
+
+  p_glEnable(myBuffer.target);
 
   // Make sure any old parts of the screen are erased
-  // Do it for both buffers!
   p_glClear(GL_COLOR_BUFFER_BIT);
   SDL_GL_SwapBuffers();
   p_glClear(GL_COLOR_BUFFER_BIT);
 
-  myOSystem->eventHandler().refreshDisplay();
-
   return true;
 }
 
@@ -323,62 +334,56 @@ void FrameBufferGL::drawMediaSource()
   uInt16* buffer       = (uInt16*) myTexture->pixels;
 
   // TODO - is this fast enough?
-  switch((int)myUsePhosphor) // use switch/case, since we'll eventually have filters
+  if(!myUsePhosphor)
   {
-    case 0:
+    uInt32 bufofsY    = 0;
+    uInt32 screenofsY = 0;
+    for(uInt32 y = 0; y < height; ++y )
     {
-      uInt32 bufofsY    = 0;
-      uInt32 screenofsY = 0;
-      for(uInt32 y = 0; y < height; ++y )
+      uInt32 pos = screenofsY;
+      for(uInt32 x = 0; x < width; ++x )
       {
-        uInt32 pos = screenofsY;
-        for(uInt32 x = 0; x < width; ++x )
+        const uInt32 bufofs = bufofsY + x;
+        uInt8 v = currentFrame[bufofs];
+        uInt8 w = previousFrame[bufofs];
+
+        if(v != w || theRedrawTIAIndicator)
         {
-          const uInt32 bufofs = bufofsY + x;
-          uInt8 v = currentFrame[bufofs];
-          uInt8 w = previousFrame[bufofs];
+          // If we ever get to this point, we know the current and previous
+          // buffers differ.  In that case, make sure the changes are
+          // are drawn in postFrameUpdate()
+          myDirtyFlag = true;
 
-          if(v != w || theRedrawTIAIndicator)
-          {
-            // If we ever get to this point, we know the current and previous
-            // buffers differ.  In that case, make sure the changes are
-            // are drawn in postFrameUpdate()
-            myDirtyFlag = true;
-
-            buffer[pos] = buffer[pos+1] = (uInt16) myPalette[v];
-          }
-          pos += 2;
+          buffer[pos] = buffer[pos+1] = (uInt16) myDefPalette[v];
         }
-        bufofsY    += width;
-        screenofsY += myTexture->w;
+        pos += 2;
       }
-      break;
+      bufofsY    += width;
+      screenofsY += myBuffer.pitch;
     }
+  }
+  else
+  {
+    // Phosphor mode always implies a dirty update,
+    // so we don't care about theRedrawTIAIndicator
+    myDirtyFlag = true;
 
-    case 1:
+    uInt32 bufofsY    = 0;
+    uInt32 screenofsY = 0;
+    for(uInt32 y = 0; y < height; ++y )
     {
-      // Phosphor mode always implies a dirty update,
-      // so we don't care about theRedrawTIAIndicator
-      myDirtyFlag = true;
-
-      uInt32 bufofsY    = 0;
-      uInt32 screenofsY = 0;
-      for(uInt32 y = 0; y < height; ++y )
+      uInt32 pos = screenofsY;
+      for(uInt32 x = 0; x < width; ++x )
       {
-        uInt32 pos = screenofsY;
-        for(uInt32 x = 0; x < width; ++x )
-        {
-          const uInt32 bufofs = bufofsY + x;
-          uInt8 v = currentFrame[bufofs];
-          uInt8 w = previousFrame[bufofs];
+        const uInt32 bufofs = bufofsY + x;
+        uInt8 v = currentFrame[bufofs];
+        uInt8 w = previousFrame[bufofs];
 
-          buffer[pos++] = (uInt16) myAvgPalette[v][w];
-          buffer[pos++] = (uInt16) myAvgPalette[v][w];
-        }
-        bufofsY    += width;
-        screenofsY += myTexture->w;
+        buffer[pos++] = (uInt16) myAvgPalette[v][w];
+        buffer[pos++] = (uInt16) myAvgPalette[v][w];
       }
-      break;
+      bufofsY    += width;
+      screenofsY += myBuffer.pitch;
     }
   }
 }
@@ -395,16 +400,16 @@ void FrameBufferGL::postFrameUpdate()
   {
     // Texturemap complete texture to surface so we have free scaling 
     // and antialiasing 
-    uInt32 w = myBaseDim.w, h = myBaseDim.h;
+    uInt32 w = myBuffer.width, h = myBuffer.height;
 
-    p_glBindTexture(GL_TEXTURE_2D, myTextureID);
-    p_glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, myTexture->w, myTexture->h,
-                      GL_RGB, GL_UNSIGNED_SHORT_5_6_5, myTexture->pixels);
+    p_glTexSubImage2D(myBuffer.target, 0, 0, 0,
+                      myBuffer.texture_width, myBuffer.texture_height,
+                      myBuffer.format, myBuffer.type, myBuffer.pixels);
     p_glBegin(GL_QUADS);
-      p_glTexCoord2f(myTexCoord[0], myTexCoord[1]); p_glVertex2i(0, 0);
-      p_glTexCoord2f(myTexCoord[2], myTexCoord[1]); p_glVertex2i(w, 0);
-      p_glTexCoord2f(myTexCoord[2], myTexCoord[3]); p_glVertex2i(w, h);
-      p_glTexCoord2f(myTexCoord[0], myTexCoord[3]); p_glVertex2i(0, h);
+      p_glTexCoord2f(myBuffer.tex_coord[0], myBuffer.tex_coord[1]); p_glVertex2i(0, 0);
+      p_glTexCoord2f(myBuffer.tex_coord[2], myBuffer.tex_coord[1]); p_glVertex2i(w, 0);
+      p_glTexCoord2f(myBuffer.tex_coord[2], myBuffer.tex_coord[3]); p_glVertex2i(w, h);
+      p_glTexCoord2f(myBuffer.tex_coord[0], myBuffer.tex_coord[3]); p_glVertex2i(0, h);
     p_glEnd();
 
     // Now show all changes made to the texture
@@ -428,68 +433,70 @@ void FrameBufferGL::scanline(uInt32 row, uInt8* data)
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void FrameBufferGL::toggleFilter()
 {
-  if(myFilterParam == GL_NEAREST)
+  if(myBuffer.filter == GL_NEAREST)
   {
-    myFilterParam = GL_LINEAR;
+    myBuffer.filter = GL_LINEAR;
     myOSystem->settings().setString("gl_filter", "linear");
     showMessage("Filtering: GL_LINEAR");
   }
   else
   {
-    myFilterParam = GL_NEAREST;
+    myBuffer.filter = GL_NEAREST;
     myOSystem->settings().setString("gl_filter", "nearest");
     showMessage("Filtering: GL_NEAREST");
   }
 
-  p_glBindTexture(GL_TEXTURE_2D, myTextureID);
-  p_glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
-  p_glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
-  p_glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, myFilterParam);
-  p_glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, myFilterParam);
+  p_glBindTexture(myBuffer.target, myBuffer.texture);
+  p_glTexParameteri(myBuffer.target, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
+  p_glTexParameteri(myBuffer.target, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
+  p_glTexParameteri(myBuffer.target, GL_TEXTURE_MAG_FILTER, myBuffer.filter);
+  p_glTexParameteri(myBuffer.target, GL_TEXTURE_MIN_FILTER, myBuffer.filter);
 
   // The filtering has changed, so redraw the entire screen
   theRedrawTIAIndicator = true;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void FrameBufferGL::hLine(uInt32 x, uInt32 y, uInt32 x2, OverlayColor color)
+void FrameBufferGL::hLine(uInt32 x, uInt32 y, uInt32 x2, int color)
 {
-  uInt16* buffer = (uInt16*) myTexture->pixels + y * myTexture->w + x;
-  while(x++ <= x2)
-    *buffer++ = (uInt16) myPalette[color];
+  // Horizontal line
+  SDL_Rect tmp;
+  tmp.x = x;
+  tmp.y = y;
+  tmp.w = x2 - x + 1;
+  tmp.h = 1;
+  SDL_FillRect(myTexture, &tmp, myDefPalette[color]);
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void FrameBufferGL::vLine(uInt32 x, uInt32 y, uInt32 y2, OverlayColor color)
+void FrameBufferGL::vLine(uInt32 x, uInt32 y, uInt32 y2, int color)
 {
-  uInt16* buffer = (uInt16*) myTexture->pixels + y * myTexture->w + x;
-  while(y++ <= y2)
-  {
-    *buffer = (uInt16) myPalette[color];
-    buffer += myTexture->w;
-  }
+  // Vertical line
+  SDL_Rect tmp;
+  tmp.x = x;
+  tmp.y = y;
+  tmp.w = 1;
+  tmp.h = y2 - y + 1;
+  SDL_FillRect(myTexture, &tmp, myDefPalette[color]);
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void FrameBufferGL::fillRect(uInt32 x, uInt32 y, uInt32 w, uInt32 h,
-                             OverlayColor color)
+                             int color)
 {
-  SDL_Rect tmp;
-
   // Fill the rectangle
+  SDL_Rect tmp;
   tmp.x = x;
   tmp.y = y;
   tmp.w = w;
   tmp.h = h;
-  SDL_FillRect(myTexture, &tmp, myPalette[color]);
+  SDL_FillRect(myTexture, &tmp, myDefPalette[color]);
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void FrameBufferGL::drawChar(const GUI::Font* FONT, uInt8 chr,
-                             uInt32 tx, uInt32 ty, OverlayColor color)
+void FrameBufferGL::drawChar(const GUI::Font* font, uInt8 chr,
+                             uInt32 tx, uInt32 ty, int color)
 {
-  GUI::Font* font = (GUI::Font*) FONT;
-
   // If this character is not included in the font, use the default char.
   if(chr < font->desc().firstchar ||
      chr >= font->desc().firstchar + font->desc().size)
@@ -505,36 +512,34 @@ void FrameBufferGL::drawChar(const GUI::Font* FONT, uInt8 chr,
   const uInt16* tmp = font->desc().bits + (font->desc().offset ?
                       font->desc().offset[chr] : (chr * h));
 
-  uInt16* buffer = (uInt16*) myTexture->pixels + ty * myTexture->w + tx;
+  uInt16* buffer = (uInt16*) myTexture->pixels + ty * myBuffer.pitch + tx;
   for(int y = 0; y < h; ++y)
   {
     const uInt16 ptr = *tmp++;
     uInt16 mask = 0x8000;
 
     for(int x = 0; x < w; ++x, mask >>= 1)
-    {
       if(ptr & mask)
-        buffer[x] = (uInt16) myPalette[color];
-    }
-    buffer += myTexture->w;
+        buffer[x] = (uInt16) myDefPalette[color];
+
+    buffer += myBuffer.pitch;
   }
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void FrameBufferGL::drawBitmap(uInt32* bitmap, Int32 tx, Int32 ty,
-                               OverlayColor color, Int32 h)
+                               int color, Int32 h)
 {
-  uInt16* buffer = (uInt16*) myTexture->pixels + ty * myTexture->w + tx;
+  uInt16* buffer = (uInt16*) myTexture->pixels + ty * myBuffer.pitch + tx;
 
   for(int y = 0; y < h; ++y)
   {
     uInt32 mask = 0xF0000000;
     for(int x = 0; x < 8; ++x, mask >>= 4)
-    {
       if(bitmap[y] & mask)
-        buffer[x] = (uInt16) myPalette[color];
-    }
-    buffer += myTexture->w;
+        buffer[x] = (uInt16) myDefPalette[color];
+
+    buffer += myBuffer.pitch;
   }
 }
 
@@ -542,93 +547,121 @@ void FrameBufferGL::drawBitmap(uInt32* bitmap, Int32 tx, Int32 ty,
 void FrameBufferGL::translateCoords(Int32* x, Int32* y)
 {
   // Wow, what a mess :)
-  *x = (Int32) (((*x - myImageDim.x) / (theZoomLevel * myFSScaleFactor * theAspectRatio)));
-  *y = (Int32) (((*y - myImageDim.y) / (theZoomLevel * myFSScaleFactor)));
+  *x = (Int32) (((*x - myImageDim.x) / (myZoomLevel * myFSScaleFactor * theAspectRatio)));
+  *y = (Int32) (((*y - myImageDim.y) / (myZoomLevel * myFSScaleFactor)));
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void FrameBufferGL::addDirtyRect(uInt32 x, uInt32 y, uInt32 w, uInt32 h)
 {
-  // TODO
-  //    Add logic to create one large dirty-rect that encompasses all
-  //    smaller dirty rects.  Then in postFrameUpdate(), change the
-  //    coordinates to only update that portion of the texture.
   myDirtyFlag = true;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void FrameBufferGL::enablePhosphor(bool enable)
+void FrameBufferGL::enablePhosphor(bool enable, int blend)
 {
   myUsePhosphor   = enable;
-  myPhosphorBlend = myOSystem->settings().getInt("ppblend");
-}
+  myPhosphorBlend = blend;
 
-// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void FrameBufferGL::cls()
-{
-/* FIXME - commented out until I figure out why it crashes in OSX
-  if(myFuncsLoaded)
-  {
-    p_glClear(GL_COLOR_BUFFER_BIT);
-    SDL_GL_SwapBuffers();
-    p_glClear(GL_COLOR_BUFFER_BIT);
-  }
-*/
+  theRedrawTIAIndicator = true;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 bool FrameBufferGL::createTextures()
 {
-  if(myTexture)
+  if(myTexture)         SDL_FreeSurface(myTexture);
+  if(myBuffer.texture)  p_glDeleteTextures(1, &myBuffer.texture);
+  memset(&myBuffer, 0, sizeof(glBufferType));
+  myBuffer.filter = GL_NEAREST;
+
+  // Fill buffer struct with valid data
+  // This changes depending on the texturing used
+  myBuffer.width  = myBaseDim.w;
+  myBuffer.height = myBaseDim.h;
+  myBuffer.tex_coord[0] = 0.0f;
+  myBuffer.tex_coord[1] = 0.0f;
+  if(myHaveTexRectEXT)
   {
-    SDL_FreeSurface(myTexture);
-    myTexture = NULL;
+    myBuffer.texture_width  = myBuffer.width;
+    myBuffer.texture_height = myBuffer.height;
+    myBuffer.target         = GL_TEXTURE_RECTANGLE_ARB;
+    myBuffer.tex_coord[2]   = (GLfloat) myBuffer.texture_width;
+    myBuffer.tex_coord[3]   = (GLfloat) myBuffer.texture_height;
+  }
+  else
+  {
+    myBuffer.texture_width  = power_of_two(myBuffer.width);
+    myBuffer.texture_height = power_of_two(myBuffer.height);
+    myBuffer.target         = GL_TEXTURE_2D;
+    myBuffer.tex_coord[2]   = (GLfloat) myBuffer.width / myBuffer.texture_width;
+    myBuffer.tex_coord[3]   = (GLfloat) myBuffer.height / myBuffer.texture_height;
   }
 
-  p_glDeleteTextures(1, &myTextureID);
-
-  uInt32 w = power_of_two(myBaseDim.w);
-  uInt32 h = power_of_two(myBaseDim.h);
-
-  myTexCoord[0] = 0.0f;
-  myTexCoord[1] = 0.0f;
-  myTexCoord[2] = (GLfloat) myBaseDim.w / w;
-  myTexCoord[3] = (GLfloat) myBaseDim.h / h;
-
-  myTexture = SDL_CreateRGBSurface(SDL_SWSURFACE, w, h, 16,
-    0x0000F800, 0x000007E0, 0x0000001F, 0x00000000);
-
+  // Create a texture that best suits the current display depth and system
+  // This code needs to be Apple-specific, otherwise performance is
+  // terrible on a Mac Mini
+#if defined(MAC_OSX)
+  myTexture = SDL_CreateRGBSurface(SDL_SWSURFACE,
+                myBuffer.texture_width, myBuffer.texture_height, 16,
+                0x00007c00, 0x000003e0, 0x0000001f, 0x00000000);
+#else
+  myTexture = SDL_CreateRGBSurface(SDL_SWSURFACE,
+                myBuffer.texture_width, myBuffer.texture_height, 16,
+                0x0000f800, 0x000007e0, 0x0000001f, 0x00000000);
+#endif
   if(myTexture == NULL)
     return false;
 
+  myBuffer.pixels = myTexture->pixels;
+  switch(myTexture->format->BytesPerPixel)
+  {
+    case 2:  // 16-bit
+      myBuffer.pitch = myTexture->pitch/2;
+      break;
+    case 3:  // 24-bit
+      myBuffer.pitch = myTexture->pitch;
+      break;
+    case 4:  // 32-bit
+      myBuffer.pitch = myTexture->pitch/4;
+      break;
+    default:
+      break;
+  }
+
   // Create an OpenGL texture from the SDL texture
-  string filter = myOSystem->settings().getString("gl_filter");
+  const string& filter = myOSystem->settings().getString("gl_filter");
   if(filter == "linear")
   {
-    myFilterParam     = GL_LINEAR;
+    myBuffer.filter   = GL_LINEAR;
     myFilterParamName = "GL_LINEAR";
   }
   else if(filter == "nearest")
   {
-    myFilterParam     = GL_NEAREST;
+    myBuffer.filter   = GL_NEAREST;
     myFilterParamName = "GL_NEAREST";
   }
 
-  p_glGenTextures(1, &myTextureID);
-  p_glBindTexture(GL_TEXTURE_2D, myTextureID);
-  p_glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
-  p_glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
-  p_glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, myFilterParam);
-  p_glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, myFilterParam);
-  p_glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, w, h, 0, GL_RGB, GL_UNSIGNED_SHORT_5_6_5,
-               myTexture->pixels);
+  p_glGenTextures(1, &myBuffer.texture);
+  p_glBindTexture(myBuffer.target, myBuffer.texture);
+  p_glTexParameteri(myBuffer.target, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
+  p_glTexParameteri(myBuffer.target, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
+  p_glTexParameteri(myBuffer.target, GL_TEXTURE_MIN_FILTER, myBuffer.filter);
+  p_glTexParameteri(myBuffer.target, GL_TEXTURE_MAG_FILTER, myBuffer.filter);
 
-  p_glDisable(GL_DEPTH_TEST);
-  p_glDisable(GL_CULL_FACE);
-  p_glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
-  p_glEnable(GL_TEXTURE_2D);
-
-  p_glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
+  // Finally, create the texture in the most optimal format
+  GLenum tex_intformat;
+#if defined (MAC_OSX)
+  tex_intformat   = GL_RGB5;
+  myBuffer.format = GL_BGRA;
+  myBuffer.type   = GL_UNSIGNED_SHORT_1_5_5_5_REV;
+#else
+  tex_intformat   = GL_RGB;
+  myBuffer.format = GL_RGB;
+  myBuffer.type   = GL_UNSIGNED_SHORT_5_6_5;
+#endif
+  p_glTexImage2D(myBuffer.target, 0, tex_intformat,
+                 myBuffer.texture_width, myBuffer.texture_height, 0,
+                 myBuffer.format, myBuffer.type, myBuffer.pixels);
 
   return true;
 }
@@ -640,8 +673,8 @@ void FrameBufferGL::setDimensions(GLdouble* orthoWidth, GLdouble* orthoHeight)
   // We have to determine final image dimensions as well as screen dimensions
   myImageDim.x = 0;
   myImageDim.y = 0;
-  myImageDim.w = (Uint16) (myBaseDim.w * theZoomLevel * theAspectRatio);
-  myImageDim.h = (Uint16) (myBaseDim.h * theZoomLevel);
+  myImageDim.w = (Uint16) (myBaseDim.w * myZoomLevel * theAspectRatio);
+  myImageDim.h = (Uint16) (myBaseDim.h * myZoomLevel);
   myScreenDim  = myImageDim;
 
   myFSScaleFactor = 1.0f;
@@ -654,8 +687,13 @@ void FrameBufferGL::setDimensions(GLdouble* orthoWidth, GLdouble* orthoHeight)
     float scaleX = 0.0f;
     float scaleY = 0.0f;
 
-    if(myOSystem->settings().getBool("gl_fsmax") &&
-       myDesktopDim.w != 0 && myDesktopDim.h != 0)
+    // When in a fullscreen, most text mode, attempt to do 'nice' scaling
+    // This means scaling will only use integral values (1, 2, 3, ...)
+    bool gl_fsmax = myOSystem->settings().getBool("gl_fsmax");
+    bool nicescale = myOSystem->eventHandler().state() == EventHandler::S_LAUNCHER ||
+                     myOSystem->eventHandler().state() == EventHandler::S_DEBUGGER;
+
+    if(gl_fsmax && myDesktopDim.w != 0 && myDesktopDim.h != 0)
     {
       // Use the largest available screen size
       myScreenDim.w = myDesktopDim.w;
@@ -669,11 +707,13 @@ void FrameBufferGL::setDimensions(GLdouble* orthoWidth, GLdouble* orthoHeight)
       else
         myFSScaleFactor = float(myScreenDim.h) / myImageDim.h;
 
+      if(nicescale)
+        myFSScaleFactor = (int)myFSScaleFactor;
+
       myImageDim.w = (Uint16) (myFSScaleFactor * myImageDim.w);
       myImageDim.h = (Uint16) (myFSScaleFactor * myImageDim.h);
     }
-    else if(myOSystem->settings().getBool("gl_fsmax") &&
-            myScreenmode != (SDL_Rect**) -1)
+    else if(gl_fsmax && myScreenmode != (SDL_Rect**) -1)
     {
       // Use the largest available screen size
       myScreenDim.w = myScreenmode[0]->w;
@@ -687,6 +727,9 @@ void FrameBufferGL::setDimensions(GLdouble* orthoWidth, GLdouble* orthoHeight)
       else
         myFSScaleFactor = float(myScreenDim.h) / myImageDim.h;
 
+      if(nicescale)
+        myFSScaleFactor = (int)myFSScaleFactor;
+
       myImageDim.w = (Uint16) (myFSScaleFactor * myImageDim.w);
       myImageDim.h = (Uint16) (myFSScaleFactor * myImageDim.h);
     }
@@ -713,13 +756,13 @@ void FrameBufferGL::setDimensions(GLdouble* orthoWidth, GLdouble* orthoHeight)
     myImageDim.x = (myScreenDim.w - myImageDim.w) / 2;
     myImageDim.y = (myScreenDim.h - myImageDim.h) / 2;
 
-    *orthoWidth  = (GLdouble) (myImageDim.w / (theZoomLevel * theAspectRatio * myFSScaleFactor));
-    *orthoHeight = (GLdouble) (myImageDim.h / (theZoomLevel * myFSScaleFactor));
+    *orthoWidth  = (GLdouble) (myImageDim.w / (myZoomLevel * theAspectRatio * myFSScaleFactor));
+    *orthoHeight = (GLdouble) (myImageDim.h / (myZoomLevel * myFSScaleFactor));
   }
   else
   {
-    *orthoWidth  = (GLdouble) (myImageDim.w / (theZoomLevel * theAspectRatio));
-    *orthoHeight = (GLdouble) (myImageDim.h / theZoomLevel);
+    *orthoWidth  = (GLdouble) (myImageDim.w / (myZoomLevel * theAspectRatio));
+    *orthoHeight = (GLdouble) (myImageDim.h / myZoomLevel);
   }
 /*
   cerr << "myImageDim.x = " << myImageDim.x << ", myImageDim.y = " << myImageDim.y << endl;
