@@ -8,33 +8,34 @@
 //  SS  SS   tt   ee      ll   ll  aa  aa
 //   SSSS     ttt  eeeee llll llll  aaaaa
 //
-// Copyright (c) 1995-2005 by Bradford W. Mott and the Stella team
+// Copyright (c) 1995-2007 by Bradford W. Mott and the Stella team
 //
 // See the file "license" for information on usage and redistribution of
 // this file, and for a DISCLAIMER OF ALL WARRANTIES.
 //
-// $Id: FrameBufferSoft.cxx,v 1.45 2006/01/20 20:07:25 stephena Exp $
+// $Id: FrameBufferSoft.cxx,v 1.69 2007/01/15 00:07:51 stephena Exp $
 //============================================================================
 
-#include <SDL.h>
-#include <SDL_syswm.h>
 #include <sstream>
+#include <SDL.h>
 
 #include "Console.hxx"
-#include "FrameBuffer.hxx"
-#include "FrameBufferSoft.hxx"
 #include "MediaSrc.hxx"
 #include "Settings.hxx"
 #include "OSystem.hxx"
 #include "Font.hxx"
 #include "GuiUtils.hxx"
+#include "RectList.hxx"
+#include "FrameBufferSoft.hxx"
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 FrameBufferSoft::FrameBufferSoft(OSystem* osystem)
   : FrameBuffer(osystem),
+    myZoomLevel(1),
+    myRenderType(kDirtyRect),
+    myUseDirtyRects(true),
     myRectList(NULL),
-    myOverlayRectList(NULL),
-    myRenderType(kSoftZoom)
+    myOverlayRectList(NULL)
 {
 }
 
@@ -61,14 +62,27 @@ bool FrameBufferSoft::initSubsystem()
   }
 
   // Create the screen
-  if(!createScreen())
-    return false;
+  return createScreen();
+}
 
-  // Show some info
-  if(myOSystem->settings().getBool("showinfo"))
-    cout << "Video rendering: Software mode" << endl << endl;
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+string FrameBufferSoft::about()
+{
+  ostringstream buf;
 
-  return true;
+  buf << "Video rendering: Software mode" << endl
+      << "  Color: " << (int)myFormat->BitsPerPixel << " bit" << endl
+      << "  Rmask = " << hex << setw(4) << (int)myFormat->Rmask
+      << ", Rshift = "<< dec << setw(2) << (int)myFormat->Rshift
+      << ", Rloss = " << dec << setw(2) << (int)myFormat->Rloss << endl
+      << "  Gmask = " << hex << setw(4) << (int)myFormat->Gmask
+      << ", Gshift = "<< dec << setw(2) << (int)myFormat->Gshift
+      << ", Gloss = " << dec << setw(2) << (int)myFormat->Gloss << endl
+      << "  Bmask = " << hex << setw(4) << (int)myFormat->Bmask
+      << ", Bshift = "<< dec << setw(2) << (int)myFormat->Bshift
+      << ", Bloss = " << dec << setw(2) << (int)myFormat->Bloss << endl;
+
+  return buf.str();
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -78,13 +92,33 @@ void FrameBufferSoft::setAspectRatio()
   theAspectRatio = 1.0;
 }
 
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+void FrameBufferSoft::setScaler(Scaler scaler)
+{
+  switch(scaler.type)
+  {
+    case kZOOM1X:
+    case kZOOM2X:
+    case kZOOM3X:
+    case kZOOM4X:
+    case kZOOM5X:
+    case kZOOM6X:
+      // Software framebuffer doesn't handle the fancy scaling modes
+      myZoomLevel = scaler.zoom;
+      break;
+    default:  // should never get here
+      myZoomLevel = 1;
+      break;
+  }
+}
+
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 bool FrameBufferSoft::createScreen()
 {
   myScreenDim.x = myScreenDim.y = 0;
 
-  myScreenDim.w = myBaseDim.w * theZoomLevel;
-  myScreenDim.h = myBaseDim.h * theZoomLevel;
+  myScreenDim.w = myBaseDim.w * myZoomLevel;
+  myScreenDim.h = myBaseDim.h * myZoomLevel;
 
   // In software mode, the image and screen dimensions are always the same
   myImageDim = myScreenDim;
@@ -95,19 +129,16 @@ bool FrameBufferSoft::createScreen()
     cerr << "ERROR: Unable to open SDL window: " << SDL_GetError() << endl;
     return false;
   }
-  switch(myScreen->format->BitsPerPixel)
-  {
-    case 16:
-      myPitch = myScreen->pitch/2;
-      break;
-    case 24:
-      myPitch = myScreen->pitch;
-      break;
-    case 32:
-      myPitch = myScreen->pitch/4;
-      break;
-  }
-  myOSystem->eventHandler().refreshDisplay();
+  myBytesPerPixel = myScreen->format->BytesPerPixel;
+  myFormat = myScreen->format;
+
+  // Make sure drawMediaSource() knows which renderer to use
+  stateChanged(myOSystem->eventHandler().state());
+
+  // Erase old rects, since they've probably been scaled for
+  // a different sized screen
+  myRectList->start();
+  myOverlayRectList->start();
 
   return true;
 }
@@ -119,14 +150,14 @@ void FrameBufferSoft::drawMediaSource()
 
   uInt8* currentFrame   = mediasrc.currentFrameBuffer();
   uInt8* previousFrame  = mediasrc.previousFrameBuffer();
-  uInt16 screenMultiple = (uInt16) theZoomLevel;
+  uInt16 screenMultiple = (uInt16) myZoomLevel;
 
   uInt32 width  = mediasrc.width();
   uInt32 height = mediasrc.height();
 
-  switch((int)myRenderType) // use switch/case, since we'll eventually have filters
+  switch(myRenderType) // use switch/case, since we'll eventually have filters
   {
-    case kSoftZoom:
+    case kDirtyRect:
     {
       struct Rectangle
       {
@@ -221,7 +252,7 @@ void FrameBufferSoft::drawMediaSource()
             temp.h = active.height * screenMultiple;
 
             myRectList->add(&temp);
-            SDL_FillRect(myScreen, &temp, myPalette[active.color]);
+            SDL_FillRect(myScreen, &temp, myDefPalette[active.color]);
 
             ++activeIndex;
           }
@@ -239,7 +270,7 @@ void FrameBufferSoft::drawMediaSource()
           temp.h = active.height * screenMultiple;
 
           myRectList->add(&temp);
-          SDL_FillRect(myScreen, &temp, myPalette[active.color]);
+          SDL_FillRect(myScreen, &temp, myDefPalette[active.color]);
         }
 
         // We can now make the current rectangles into the active rectangles
@@ -264,33 +295,165 @@ void FrameBufferSoft::drawMediaSource()
         temp.h = active.height * screenMultiple;
 
         myRectList->add(&temp);
-        SDL_FillRect(myScreen, &temp, myPalette[active.color]);
+        SDL_FillRect(myScreen, &temp, myDefPalette[active.color]);
       }
-      break; // case 0
+      break;  // kDirtyRect
     }
 
-    case kPhosphor_16:
+    case kSoftZoom_16:
     {
-      // Since phosphor mode updates the whole screen,
-      // we might as well use SDL_Flip (see postFrameUpdate)
-      myUseDirtyRects = false;
       SDL_Rect temp;
       temp.x = temp.y = temp.w = temp.h = 0;
       myRectList->add(&temp);
 
+      SDL_LockSurface(myScreen);
       uInt16* buffer    = (uInt16*)myScreen->pixels;
       uInt32 bufofsY    = 0;
       uInt32 screenofsY = 0;
-      for(uInt32 y = 0; y < height; ++y )
+      for(uInt32 y = 0; y < height; ++y)
       {
-        uInt32 ystride = theZoomLevel;
+        uInt32 ystride = myZoomLevel;
         while(ystride--)
         {
           uInt32 pos = screenofsY;
-          for(uInt32 x = 0; x < width; ++x )
+          for(uInt32 x = 0; x < width; ++x)
           {
             const uInt32 bufofs = bufofsY + x;
-            uInt32 xstride = theZoomLevel;
+            uInt32 xstride = myZoomLevel;
+
+            uInt8 v = currentFrame[bufofs];
+            uInt8 w = previousFrame[bufofs];
+
+            if(v != w || theRedrawTIAIndicator)
+            {
+              while(xstride--)
+              {
+                buffer[pos++] = (uInt16) myDefPalette[v];
+                buffer[pos++] = (uInt16) myDefPalette[v];
+              }
+            }
+            else
+              pos += xstride + xstride;
+          }
+          screenofsY += myPitch;
+        }
+        bufofsY += width;
+      }
+      SDL_UnlockSurface(myScreen);
+      break;  // kSoftZoom_16
+    }
+
+    case kSoftZoom_24:
+    {
+      SDL_Rect temp;
+      temp.x = temp.y = temp.w = temp.h = 0;
+      myRectList->add(&temp);
+
+      SDL_LockSurface(myScreen);
+      uInt8* buffer     = (uInt8*)myScreen->pixels;
+      uInt32 bufofsY    = 0;
+      uInt32 screenofsY = 0;
+      for(uInt32 y = 0; y < height; ++y)
+      {
+        uInt32 ystride = myZoomLevel;
+        while(ystride--)
+        {
+          uInt32 pos = screenofsY;
+          for(uInt32 x = 0; x < width; ++x)
+          {
+            const uInt32 bufofs = bufofsY + x;
+            uInt32 xstride = myZoomLevel;
+
+            uInt8 v = currentFrame[bufofs];
+            uInt8 w = previousFrame[bufofs];
+
+            if(v != w || theRedrawTIAIndicator)
+            {
+              uInt32 pixel = myDefPalette[v];
+              uInt8 r = (pixel & myFormat->Rmask) >> myFormat->Rshift;
+              uInt8 g = (pixel & myFormat->Gmask) >> myFormat->Gshift;
+              uInt8 b = (pixel & myFormat->Bmask) >> myFormat->Bshift;
+
+              while(xstride--)
+              {
+                buffer[pos++] = r;  buffer[pos++] = g;  buffer[pos++] = b;
+                buffer[pos++] = r;  buffer[pos++] = g;  buffer[pos++] = b;
+              }
+            }
+            else  // try to eliminate multply whereever possible
+              pos += xstride + xstride + xstride + xstride + xstride + xstride;
+          }
+          screenofsY += myPitch;
+        }
+        bufofsY += width;
+      }
+      SDL_UnlockSurface(myScreen);
+      break;  // kSoftZoom_24
+    }
+
+    case kSoftZoom_32:
+    {
+      SDL_Rect temp;
+      temp.x = temp.y = temp.w = temp.h = 0;
+      myRectList->add(&temp);
+
+      SDL_LockSurface(myScreen);
+      uInt32* buffer    = (uInt32*)myScreen->pixels;
+      uInt32 bufofsY    = 0;
+      uInt32 screenofsY = 0;
+      for(uInt32 y = 0; y < height; ++y)
+      {
+        uInt32 ystride = myZoomLevel;
+        while(ystride--)
+        {
+          uInt32 pos = screenofsY;
+          for(uInt32 x = 0; x < width; ++x)
+          {
+            const uInt32 bufofs = bufofsY + x;
+            uInt32 xstride = myZoomLevel;
+
+            uInt8 v = currentFrame[bufofs];
+            uInt8 w = previousFrame[bufofs];
+
+            if(v != w || theRedrawTIAIndicator)
+            {
+              while(xstride--)
+              {
+                buffer[pos++] = (uInt32) myDefPalette[v];
+                buffer[pos++] = (uInt32) myDefPalette[v];
+              }
+            }
+            else
+              pos += xstride + xstride;
+          }
+          screenofsY += myPitch;
+        }
+        bufofsY += width;
+      }
+      SDL_UnlockSurface(myScreen);
+      break;  // kSoftZoom_32
+    }
+
+    case kPhosphor_16:
+    {
+      SDL_Rect temp;
+      temp.x = temp.y = temp.w = temp.h = 0;
+      myRectList->add(&temp);
+
+      SDL_LockSurface(myScreen);
+      uInt16* buffer    = (uInt16*)myScreen->pixels;
+      uInt32 bufofsY    = 0;
+      uInt32 screenofsY = 0;
+      for(uInt32 y = 0; y < height; ++y)
+      {
+        uInt32 ystride = myZoomLevel;
+        while(ystride--)
+        {
+          uInt32 pos = screenofsY;
+          for(uInt32 x = 0; x < width; ++x)
+          {
+            const uInt32 bufofs = bufofsY + x;
+            uInt32 xstride = myZoomLevel;
 
             uInt8 v = currentFrame[bufofs];
             uInt8 w = previousFrame[bufofs];
@@ -305,49 +468,37 @@ void FrameBufferSoft::drawMediaSource()
         }
         bufofsY += width;
       }
+      SDL_UnlockSurface(myScreen);
       break;  // kPhosphor_16
     }
 
     case kPhosphor_24:
     {
-      // Since phosphor mode updates the whole screen,
-      // we might as well use SDL_Flip (see postFrameUpdate)
-      myUseDirtyRects = false;
       SDL_Rect temp;
       temp.x = temp.y = temp.w = temp.h = 0;
       myRectList->add(&temp);
 
+      SDL_LockSurface(myScreen);
       uInt8* buffer     = (uInt8*)myScreen->pixels;
       uInt32 bufofsY    = 0;
       uInt32 screenofsY = 0;
-      for(uInt32 y = 0; y < height; ++y )
+      for(uInt32 y = 0; y < height; ++y)
       {
-        uInt32 ystride = theZoomLevel;
+        uInt32 ystride = myZoomLevel;
         while(ystride--)
         {
           uInt32 pos = screenofsY;
-          for(uInt32 x = 0; x < width; ++x )
+          for(uInt32 x = 0; x < width; ++x)
           {
             const uInt32 bufofs = bufofsY + x;
-            uInt32 xstride = theZoomLevel;
+            uInt32 xstride = myZoomLevel;
 
             uInt8 v = currentFrame[bufofs];
             uInt8 w = previousFrame[bufofs];
-            uInt8 r, g, b;
-            if(SDL_BYTEORDER == SDL_BIG_ENDIAN)
-            {
-              uInt32 pixel = myAvgPalette[v][w];
-              b = pixel & 0xff;
-              g = (pixel & 0xff00) >> 8;
-              r = (pixel & 0xff0000) >> 16;
-            }
-            else
-            {
-              uInt32 pixel = myAvgPalette[v][w];
-              r = pixel & 0xff;
-              g = (pixel & 0xff00) >> 8;
-              b = (pixel & 0xff0000) >> 16;
-            }
+            uInt32 pixel = myAvgPalette[v][w];
+            uInt8 r = (pixel & myFormat->Rmask) >> myFormat->Rshift;
+            uInt8 g = (pixel & myFormat->Gmask) >> myFormat->Gshift;
+            uInt8 b = (pixel & myFormat->Bmask) >> myFormat->Bshift;
 
             while(xstride--)
             {
@@ -359,31 +510,30 @@ void FrameBufferSoft::drawMediaSource()
         }
         bufofsY += width;
       }
+      SDL_UnlockSurface(myScreen);
       break;  // kPhosphor_24
     }
 
     case kPhosphor_32:
     {
-      // Since phosphor mode updates the whole screen,
-      // we might as well use SDL_Flip (see postFrameUpdate)
-      myUseDirtyRects = false;
       SDL_Rect temp;
       temp.x = temp.y = temp.w = temp.h = 0;
       myRectList->add(&temp);
 
+      SDL_LockSurface(myScreen);
       uInt32* buffer    = (uInt32*)myScreen->pixels;
       uInt32 bufofsY    = 0;
       uInt32 screenofsY = 0;
-      for(uInt32 y = 0; y < height; ++y )
+      for(uInt32 y = 0; y < height; ++y)
       {
-        uInt32 ystride = theZoomLevel;
+        uInt32 ystride = myZoomLevel;
         while(ystride--)
         {
           uInt32 pos = screenofsY;
-          for(uInt32 x = 0; x < width; ++x )
+          for(uInt32 x = 0; x < width; ++x)
           {
             const uInt32 bufofs = bufofsY + x;
-            uInt32 xstride = theZoomLevel;
+            uInt32 xstride = myZoomLevel;
 
             uInt8 v = currentFrame[bufofs];
             uInt8 w = previousFrame[bufofs];
@@ -398,6 +548,7 @@ void FrameBufferSoft::drawMediaSource()
         }
         bufofsY += width;
       }
+      SDL_UnlockSurface(myScreen);
       break;  // kPhosphor_32
     }
   }
@@ -419,9 +570,6 @@ void FrameBufferSoft::preFrameUpdate()
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void FrameBufferSoft::postFrameUpdate()
 {
-  // This is a performance hack until I have more time to work
-  // on the Win32 code.  It seems that SDL_UpdateRects() is very
-  // expensive in Windows, so we force a full screen update instead.
   if(myUseDirtyRects)
     SDL_UpdateRects(myScreen, myRectList->numRects(), myRectList->rects());
   else if(myRectList->numRects() > 0)
@@ -437,19 +585,16 @@ void FrameBufferSoft::scanline(uInt32 row, uInt8* data)
   // Make sure no pixels are being modified
   SDL_LockSurface(myScreen);
 
-  uInt32 bpp     = myScreen->format->BytesPerPixel;
-  uInt8* start   = (uInt8*) myScreen->pixels;
-  uInt32 yoffset = row * myScreen->pitch;
   uInt32 pixel = 0;
   uInt8 *p, r, g, b;
 
   for(Int32 x = 0; x < myScreen->w; x++)
   {
-    p = (Uint8*) (start    +  // Start at top of RAM
-                 (yoffset) +  // Go down 'row' lines
-                 (x * bpp));  // Go in 'x' pixels
+    p = (Uint8*) ((uInt8*)myScreen->pixels +  // Start at top of RAM
+                 (row * myScreen->pitch) +    // Go down 'row' lines
+                 (x * myBytesPerPixel));      // Go in 'x' pixels
 
-    switch(bpp)
+    switch(myBytesPerPixel)
     {
       case 1:
         pixel = *p;
@@ -488,53 +633,51 @@ void FrameBufferSoft::toggleFilter()
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void FrameBufferSoft::hLine(uInt32 x, uInt32 y, uInt32 x2, OverlayColor color)
+void FrameBufferSoft::hLine(uInt32 x, uInt32 y, uInt32 x2, int color)
 {
   SDL_Rect tmp;
 
   // Horizontal line
-  tmp.x = x * theZoomLevel;
-  tmp.y = y * theZoomLevel;
-  tmp.w = (x2 - x + 1) * theZoomLevel;
-  tmp.h = theZoomLevel;
-  SDL_FillRect(myScreen, &tmp, myPalette[color]);
+  tmp.x = x * myZoomLevel;
+  tmp.y = y * myZoomLevel;
+  tmp.w = (x2 - x + 1) * myZoomLevel;
+  tmp.h = myZoomLevel;
+  SDL_FillRect(myScreen, &tmp, myDefPalette[color]);
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void FrameBufferSoft::vLine(uInt32 x, uInt32 y, uInt32 y2, OverlayColor color)
+void FrameBufferSoft::vLine(uInt32 x, uInt32 y, uInt32 y2, int color)
 {
   SDL_Rect tmp;
 
   // Vertical line
-  tmp.x = x * theZoomLevel;
-  tmp.y = y * theZoomLevel;
-  tmp.w = theZoomLevel;
-  tmp.h = (y2 - y + 1) * theZoomLevel;
-  SDL_FillRect(myScreen, &tmp, myPalette[color]);
+  tmp.x = x * myZoomLevel;
+  tmp.y = y * myZoomLevel;
+  tmp.w = myZoomLevel;
+  tmp.h = (y2 - y + 1) * myZoomLevel;
+  SDL_FillRect(myScreen, &tmp, myDefPalette[color]);
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void FrameBufferSoft::fillRect(uInt32 x, uInt32 y, uInt32 w, uInt32 h,
-                               OverlayColor color)
+                               int color)
 {
   SDL_Rect tmp;
 
   // Fill the rectangle
-  tmp.x = x * theZoomLevel;
-  tmp.y = y * theZoomLevel;
-  tmp.w = w * theZoomLevel;
-  tmp.h = h * theZoomLevel;
-  SDL_FillRect(myScreen, &tmp, myPalette[color]);
+  tmp.x = x * myZoomLevel;
+  tmp.y = y * myZoomLevel;
+  tmp.w = w * myZoomLevel;
+  tmp.h = h * myZoomLevel;
+  SDL_FillRect(myScreen, &tmp, myDefPalette[color]);
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void FrameBufferSoft::drawChar(const GUI::Font* FONT, uInt8 chr,
-                               uInt32 xorig, uInt32 yorig, OverlayColor color)
+void FrameBufferSoft::drawChar(const GUI::Font* font, uInt8 chr,
+                               uInt32 xorig, uInt32 yorig, int color)
 {
-  GUI::Font* font = (GUI::Font*)FONT;
-  const FontDesc& desc = font->desc();
-
   // If this character is not included in the font, use the default char.
+  const FontDesc& desc = font->desc();
   if(chr < desc.firstchar || chr >= desc.firstchar + desc.size)
   {
     if (chr == ' ')
@@ -547,28 +690,111 @@ void FrameBufferSoft::drawChar(const GUI::Font* FONT, uInt8 chr,
   chr -= desc.firstchar;
   const uInt16* tmp = desc.bits + (desc.offset ? desc.offset[chr] : (chr * h));
 
-  SDL_Rect rect;
-  for(int y = 0; y < h; y++)
-  {
-    const uInt16 buffer = *tmp++;
-    uInt16 mask = 0x8000;
+  // Scale the origins to the current zoom
+  xorig *= myZoomLevel;
+  yorig *= myZoomLevel;
 
-    for(int x = 0; x < w; x++, mask >>= 1)
+  SDL_LockSurface(myScreen);
+
+  int screenofsY = 0;
+  switch(myBytesPerPixel)
+  {
+    case 2:
     {
-      if ((buffer & mask) != 0)
+      // Get buffer position where upper-left pixel of the character will be drawn
+      uInt16* buffer = (uInt16*) myScreen->pixels + yorig * myPitch + xorig;
+      for(int y = h; y; --y)
       {
-        rect.x = (x + xorig) * theZoomLevel;
-        rect.y = (y + yorig) * theZoomLevel;
-        rect.w = rect.h = theZoomLevel;
-        SDL_FillRect(myScreen, &rect, myPalette[color]);
+        const uInt16 fontbuf = *tmp++;
+        int ystride = myZoomLevel;
+        while(ystride--)
+        {
+          uInt16 mask = 0x8000;
+          int pos = screenofsY;
+          for(int x = 0; x < w; x++, mask >>= 1)
+          {
+            int xstride = myZoomLevel;
+            if((fontbuf & mask) != 0)
+              while(xstride--)
+                buffer[pos++] = myDefPalette[color];
+            else
+              pos += xstride;
+          }
+          screenofsY += myPitch;
+        }
       }
+      break;
     }
+    case 3:
+    {
+      // Get buffer position where upper-left pixel of the character will be drawn
+      uInt8* buffer = (uInt8*) myScreen->pixels + yorig * myPitch + xorig;
+      uInt32 pixel = myDefPalette[color];
+      uInt8 r = (pixel & myFormat->Rmask) >> myFormat->Rshift;
+      uInt8 g = (pixel & myFormat->Gmask) >> myFormat->Gshift;
+      uInt8 b = (pixel & myFormat->Bmask) >> myFormat->Bshift;
+
+      for(int y = h; y; --y)
+      {
+        const uInt16 fontbuf = *tmp++;
+        int ystride = myZoomLevel;
+        while(ystride--)
+        {
+          uInt16 mask = 0x8000;
+          int pos = screenofsY;
+          for(int x = 0; x < w; x++, mask >>= 1)
+          {
+            int xstride = myZoomLevel;
+            if((fontbuf & mask) != 0)
+            {
+              while(xstride--)
+              {
+                buffer[pos++] = r;  buffer[pos++] = g;  buffer[pos++] = b;
+              }
+            }
+            else
+              pos += xstride + xstride + xstride;
+          }
+          screenofsY += myPitch;
+        }
+      }
+      break;
+    }
+    case 4:
+    {
+      // Get buffer position where upper-left pixel of the character will be drawn
+      uInt32* buffer = (uInt32*) myScreen->pixels + yorig * myPitch + xorig;
+      for(int y = h; y; --y)
+      {
+        const uInt16 fontbuf = *tmp++;
+        int ystride = myZoomLevel;
+        while(ystride--)
+        {
+          uInt16 mask = 0x8000;
+          int pos = screenofsY;
+          for(int x = 0; x < w; x++, mask >>= 1)
+          {
+            int xstride = myZoomLevel;
+            if((fontbuf & mask) != 0)
+              while(xstride--)
+                buffer[pos++] = myDefPalette[color];
+            else
+              pos += xstride;
+          }
+          screenofsY += myPitch;
+        }
+      }
+      break;
+    }
+    default:
+      break;
   }
+  SDL_UnlockSurface(myScreen);
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void FrameBufferSoft::drawBitmap(uInt32* bitmap, Int32 xorig, Int32 yorig,
-                                 OverlayColor color, Int32 h)
+                                 int color, Int32 h)
 {
   SDL_Rect rect;
   for(int y = 0; y < h; y++)
@@ -579,10 +805,10 @@ void FrameBufferSoft::drawBitmap(uInt32* bitmap, Int32 xorig, Int32 yorig,
     {
       if(bitmap[y] & mask)
       {
-        rect.x = (x + xorig) * theZoomLevel;
-        rect.y = (y + yorig) * theZoomLevel;
-        rect.w = rect.h = theZoomLevel;
-        SDL_FillRect(myScreen, &rect, myPalette[color]);
+        rect.x = (x + xorig) * myZoomLevel;
+        rect.y = (y + yorig) * myZoomLevel;
+        rect.w = rect.h = myZoomLevel;
+        SDL_FillRect(myScreen, &rect, myDefPalette[color]);
       }
     }
   }
@@ -593,35 +819,21 @@ void FrameBufferSoft::translateCoords(Int32* x, Int32* y)
 {
   // We don't bother checking offsets or aspect ratios, since
   // they're not yet supported in software mode.
-  *x /= theZoomLevel;
-  *y /= theZoomLevel;
+  *x /= myZoomLevel;
+  *y /= myZoomLevel;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void FrameBufferSoft::addDirtyRect(uInt32 x, uInt32 y, uInt32 w, uInt32 h)
 {
-  x *= theZoomLevel;
-  y *= theZoomLevel;
-  w *= theZoomLevel;
-  h *= theZoomLevel;
-
-  // Check if rect is in screen area
-  // This is probably a bug, since the GUI code shouldn't be setting
-  // a dirty rect larger than the screen
-  int x1 = x, y1 = y, x2 = x + w, y2 = y + h;
-  int sx1 = myScreenDim.x, sy1 = myScreenDim.y,
-      sx2 = myScreenDim.x + myScreenDim.w, sy2 = myScreenDim.y + myScreenDim.h;
-  if(x1 < sx1 || y1 < sy1 || x2 > sx2 || y2 > sy2)
-    return;
-
   // Add a dirty rect to the overlay rectangle list
   // They will actually be added to the main rectlist in preFrameUpdate()
   // TODO - intelligent merging of rectangles, to avoid overlap
   SDL_Rect temp;
-  temp.x = x;
-  temp.y = y;
-  temp.w = w;
-  temp.h = h;
+  temp.x = x * myZoomLevel;
+  temp.y = y * myZoomLevel;
+  temp.w = w * myZoomLevel;
+  temp.h = h * myZoomLevel;
 
   myOverlayRectList->add(&temp);
 
@@ -630,101 +842,73 @@ void FrameBufferSoft::addDirtyRect(uInt32 x, uInt32 y, uInt32 w, uInt32 h)
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void FrameBufferSoft::enablePhosphor(bool enable)
+void FrameBufferSoft::enablePhosphor(bool enable, int blend)
 {
   myUsePhosphor   = enable;
-  myPhosphorBlend = myOSystem->settings().getInt("ppblend");
+  myPhosphorBlend = blend;
 
-  if(myUsePhosphor)
+  stateChanged(myOSystem->eventHandler().state());
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+void FrameBufferSoft::stateChanged(EventHandler::State state)
+{
+  if(!myScreen)
+    return;
+
+  // When in a UI mode, always use dirty rects
+  // Otherwise, check the 'dirtyrects' setting
+  // Phosphor mode implies a full update, so turn off dirty rects
+  bool emulation = state == EventHandler::S_EMULATE;
+  if(emulation)
   {
-    switch(myScreen->format->BitsPerPixel)
-    {
-      case 16:
-        myPitch      = myScreen->pitch/2;
-        myRenderType = kPhosphor_16;
-        break;
-      case 24:
-        myPitch      = myScreen->pitch;
-        myRenderType = kPhosphor_24;
-        break;
-      case 32:
-        myPitch      = myScreen->pitch/4;
-        myRenderType = kPhosphor_32;
-        break;
-      default:
-        myRenderType = kSoftZoom; // What else should we do here?
-        break;
-    }
+    if(myUsePhosphor)
+      myUseDirtyRects = false;
+    else
+      myUseDirtyRects = myOSystem->settings().getBool("dirtyrects");
   }
   else
-    myRenderType = kSoftZoom;
-}
+    myUseDirtyRects = true;
 
-// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void FrameBufferSoft::cls()
-{
-  if(myScreen)
+  // Make sure drawMediaSource() knows which renderer to use
+  // Testing for dirty rects takes priority over phosphor mode,
+  // since phosphor mode only exists while emulating a ROM
+  switch(myBytesPerPixel)
   {
-    SDL_FillRect(myScreen, NULL, 0);
-    SDL_UpdateRect(myScreen, 0, 0, 0, 0);
-  }
-}
-
-// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-RectList::RectList(Uint32 size)
-{
-  currentSize = size;
-  currentRect = 0;
-
-  rectArray = new SDL_Rect[currentSize];
-}
-
-// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-RectList::~RectList()
-{
-  delete[] rectArray;
-}
-
-// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void RectList::add(SDL_Rect* newRect)
-{
-  if(currentRect >= currentSize)
-  {
-    currentSize = currentSize * 2;
-    SDL_Rect *temp = new SDL_Rect[currentSize];
-
-    for(Uint32 i = 0; i < currentRect; ++i)
-      temp[i] = rectArray[i];
-
-    delete[] rectArray;
-    rectArray = temp;
+    case 2:  // 16-bit
+      myPitch = myScreen->pitch/2;
+      if(myUsePhosphor && emulation)
+        myRenderType = kPhosphor_16;
+      else if(myUseDirtyRects)
+        myRenderType = kDirtyRect;
+      else
+        myRenderType = kSoftZoom_16;
+      break;
+    case 3:  // 24-bit
+      myPitch = myScreen->pitch;
+      if(myUsePhosphor && emulation)
+        myRenderType = kPhosphor_24;
+      else if(myUseDirtyRects)
+        myRenderType = kDirtyRect;
+      else
+        myRenderType = kSoftZoom_24;
+      break;
+    case 4:  // 32-bit
+      myPitch = myScreen->pitch/4;
+      if(myUsePhosphor && emulation)
+        myRenderType = kPhosphor_32;
+      else if(myUseDirtyRects)
+        myRenderType = kDirtyRect;
+      else
+        myRenderType = kSoftZoom_32;
+      break;
+    default:
+      myRenderType = kDirtyRect; // What else should we do here?
+      break;
   }
 
-//cerr << "RectList::add():  "
-//     << "x=" << newRect->x << ", y=" << newRect->y << ", w=" << newRect->w << ", h=" << newRect->h << endl;
+  // Have the changes take effect
+  myOSystem->eventHandler().refreshDisplay();
 
-  rectArray[currentRect].x = newRect->x;
-  rectArray[currentRect].y = newRect->y;
-  rectArray[currentRect].w = newRect->w;
-  rectArray[currentRect].h = newRect->h;
-
-  ++currentRect;
-}
-
-// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-SDL_Rect* RectList::rects()
-{
-  return rectArray;
-}
-
-// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-Uint32 RectList::numRects()
-{
-  return currentRect;
-}
-
-// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void RectList::start()
-{
-  currentRect = 0;
+//cerr << "Render type = " << myRenderType << ", dirty rects = " << myUseDirtyRects << endl;
 }
