@@ -8,12 +8,12 @@
 //  SS  SS   tt   ee      ll   ll  aa  aa
 //   SSSS     ttt  eeeee llll llll  aaaaa
 //
-// Copyright (c) 1995-2005 by Bradford W. Mott and the Stella team
+// Copyright (c) 1995-2007 by Bradford W. Mott and the Stella team
 //
 // See the file "license" for information on usage and redistribution of
 // this file, and for a DISCLAIMER OF ALL WARRANTIES.
 //
-// $Id: SoundSDL.cxx,v 1.28 2006/01/06 00:31:55 stephena Exp $
+// $Id: SoundSDL.cxx,v 1.36 2007/01/01 18:04:40 stephena Exp $
 //============================================================================
 
 #ifdef SOUND_SUPPORT
@@ -98,7 +98,11 @@ void SoundSDL::initialize()
       SDL_AudioSpec desired;
 #ifndef PSP
       desired.freq   = frequency;
+    #ifndef GP2X
       desired.format = AUDIO_U8;
+    #else
+      desired.format = AUDIO_U16;
+    #endif
 #else
       desired.freq   = 44100;
       desired.format = AUDIO_U16;
@@ -131,14 +135,14 @@ void SoundSDL::initialize()
       myIsMuted = false;
       myFragmentSizeLogBase2 = log((double)myHardwareSpec.samples) / log(2.0);
 
-/*
+		/*
         cerr << "Freq: " << (int)myHardwareSpec.freq << endl;
         cerr << "Format: " << (int)myHardwareSpec.format << endl;
         cerr << "Channels: " << (int)myHardwareSpec.channels << endl;
         cerr << "Silence: " << (int)myHardwareSpec.silence << endl;
         cerr << "Samples: " << (int)myHardwareSpec.samples << endl;
         cerr << "Size: " << (int)myHardwareSpec.size << endl;
-*/
+		  */
 
       // Now initialize the TIASound object which will actually generate sound
       myTIASound.outputFrequency(myHardwareSpec.freq);
@@ -158,6 +162,7 @@ void SoundSDL::initialize()
              << "  Volume     : " << myVolume << endl
              << "  Frag size  : " << fragsize << endl
              << "  Frequency  : " << myHardwareSpec.freq << endl
+             << "  Format     : " << myHardwareSpec.format << endl
              << "  TIA Freq.  : " << tiafreq << endl
              << "  Channels   : " << myNumChannels << endl
              << "  Clip volume: " << (int)clipvol << endl << endl;
@@ -213,6 +218,7 @@ void SoundSDL::reset()
     SDL_PauseAudio(1);
     myIsMuted = false;
     myLastRegisterSetCycle = 0;
+    myTIASound.reset();
     myRegWriteQueue.clear();
     SDL_PauseAudio(0);
   }
@@ -272,7 +278,7 @@ void SoundSDL::setChannels(uInt32 channels)
   if(channels == 1 || channels == 2)
   {
     myNumChannels = channels;
-    myOSystem->settings().setInt("channels", myNumChannels, false);
+    myOSystem->settings().setInt("channels", myNumChannels);
   }
 }
 
@@ -407,6 +413,26 @@ void SoundSDL::callback(void* udata, uInt8* stream, int len)
 {
   SoundSDL* sound = (SoundSDL*)udata;
   sound->processFragment(stream, (Int32)len);
+#ifdef ATARIVOX_SUPPORT
+  cerr << "SoundSDL::callback(): len==" << len << endl;
+
+  // See if we need sound from the AtariVox
+  AtariVox *vox = sound->myOSystem->console().atariVox();
+  if(vox)
+  {
+    // If so, mix 'em together (this is a crappy way to mix audio streams...)
+    uInt8 *s = stream;
+    for(int i=0; i<len/OUTPUT_BUFFER_SIZE; i++)
+    {
+      int count;
+      uInt8 *voxSamples = vox->getSpeakJet()->getSamples(&count);
+      if(!count)
+        break;
+      SDL_MixAudio(s, voxSamples, OUTPUT_BUFFER_SIZE, SDL_MIX_MAXVOLUME);
+      s += OUTPUT_BUFFER_SIZE;
+    }
+  }
+#endif
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
