@@ -8,12 +8,12 @@
 //  SS  SS   tt   ee      ll   ll  aa  aa
 //   SSSS     ttt  eeeee llll llll  aaaaa
 //
-// Copyright (c) 1995-2005 by Bradford W. Mott and the Stella team
+// Copyright (c) 1995-2007 by Bradford W. Mott and the Stella team
 //
 // See the file "license" for information on usage and redistribution of
 // this file, and for a DISCLAIMER OF ALL WARRANTIES.
 //
-// $Id: DataGridWidget.cxx,v 1.5 2006/01/15 20:46:19 stephena Exp $
+// $Id: DataGridWidget.cxx,v 1.10 2007/01/01 18:04:43 stephena Exp $
 //
 //   Based on code from ScummVM - Scumm Interpreter
 //   Copyright (C) 2002-2004 The ScummVM project
@@ -31,7 +31,8 @@
 DataGridWidget::DataGridWidget(GuiObject* boss, const GUI::Font& font,
                                int x, int y, int cols, int rows,
                                int colchars, int bits, BaseFormat base)
-  : EditableWidget(boss, x, y, cols*(colchars * font.getMaxCharWidth() + 8) + 1,
+  : EditableWidget(boss, font, x, y,
+                   cols*(colchars * font.getMaxCharWidth() + 8) + 1,
                    font.getLineHeight()*rows + 1),
     _rows(rows),
     _cols(cols),
@@ -44,9 +45,8 @@ DataGridWidget::DataGridWidget(GuiObject* boss, const GUI::Font& font,
     _selectedItem(0),
     _opsWidget(NULL)
 {
-  setFont(font);
-
-  _flags = WIDGET_ENABLED | WIDGET_CLEARBG | WIDGET_RETAIN_FOCUS;
+  _flags = WIDGET_ENABLED | WIDGET_CLEARBG | WIDGET_RETAIN_FOCUS |
+           WIDGET_WANTS_RAWDATA;
   _type = kDataGridWidget;
   _editMode = false;
 
@@ -521,8 +521,7 @@ void DataGridWidget::drawWidget(bool hilite)
         buffer = _valueStringList[pos];
         deltax = 0;
 
-        OverlayColor color = kTextColor;
-
+        int color = kTextColor;
         if(_changedList[pos])
         {
           fb.fillRect(x - 3, y - 1, _colWidth-1, _rowHeight-1, kTextColorEm);
