@@ -8,12 +8,12 @@
 //  SS  SS   tt   ee      ll   ll  aa  aa
 //   SSSS     ttt  eeeee llll llll  aaaaa
 //
-// Copyright (c) 1995-2005 by Bradford W. Mott and the Stella team
+// Copyright (c) 1995-2007 by Bradford W. Mott and the Stella team
 //
 // See the file "license" for information on usage and redistribution of
 // this file, and for a DISCLAIMER OF ALL WARRANTIES.
 //
-// $Id: PromptWidget.cxx,v 1.6 2006/01/15 20:46:19 stephena Exp $
+// $Id: PromptWidget.cxx,v 1.15 2007/01/01 18:04:44 stephena Exp $
 //
 //   Based on code from ScummVM - Scumm Interpreter
 //   Copyright (C) 2002-2004 The ScummVM project
@@ -44,18 +44,19 @@
  */
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-PromptWidget::PromptWidget(GuiObject* boss, int x, int y, int w, int h)
-  : Widget(boss, x, y, w - kScrollBarWidth, h),
+PromptWidget::PromptWidget(GuiObject* boss, const GUI::Font& font,
+                           int x, int y, int w, int h)
+  : Widget(boss, font, x, y, w - kScrollBarWidth, h),
     CommandSender(boss),
     _makeDirty(false),
     _firstTime(true)
 {
   _flags = WIDGET_ENABLED | WIDGET_CLEARBG | WIDGET_RETAIN_FOCUS |
-           WIDGET_WANTS_TAB;
+           WIDGET_WANTS_TAB | WIDGET_WANTS_RAWDATA;
   _type = kPromptWidget;
 
-  _kConsoleCharWidth  = instance()->consoleFont().getMaxCharWidth();
-  _kConsoleCharHeight = instance()->consoleFont().getFontHeight();
+  _kConsoleCharWidth  = font.getMaxCharWidth();
+  _kConsoleCharHeight = font.getFontHeight();
   _kConsoleLineHeight = _kConsoleCharHeight + 2;
 
   // Calculate depending values
@@ -70,7 +71,7 @@ PromptWidget::PromptWidget(GuiObject* boss, int x, int y, int w, int h)
   _firstLineInBuffer = 0;
 
   // Add scrollbar
-  _scrollBar = new ScrollBarWidget(boss, _x + _w, _y, kScrollBarWidth, _h);
+  _scrollBar = new ScrollBarWidget(boss, font, _x + _w, _y, kScrollBarWidth, _h);
   _scrollBar->setTarget(this);
 
   // Init colors
@@ -101,7 +102,7 @@ PromptWidget::~PromptWidget()
 void PromptWidget::drawWidget(bool hilite)
 {
 //cerr << "PromptWidget::drawWidget\n";
-  OverlayColor fgcolor, bgcolor;
+  int fgcolor, bgcolor;
 
   FrameBuffer& fb = _boss->instance()->frameBuffer();
 
@@ -117,10 +118,10 @@ void PromptWidget::drawWidget(bool hilite)
 
       if(c & (1 << 17)) { // inverse video flag
         fgcolor = bgColor;
-        bgcolor = (OverlayColor)((c & 0x1ffff) >> 8);
+        bgcolor = (c & 0x1ffff) >> 8;
         fb.fillRect(x, y, _kConsoleCharWidth, _kConsoleCharHeight, bgcolor);
       } else {
-        fgcolor = (OverlayColor)(c >> 8);
+        fgcolor = c >> 8;
         bgcolor = bgColor;
       }
       fb.drawChar(&instance()->consoleFont(), c & 0x7f, x, y, fgcolor);
@@ -178,23 +179,16 @@ bool PromptWidget::handleKeyDown(int ascii, int keycode, int modifiers)
 
       if (len > 0)
       {
-        // We have to allocate the string buffer with new, since VC++ sadly does not
-        // comply to the C++ standard, so we can't use a dynamic sized stack array.
-        char *str = new char[len + 1];
-
-        // Copy the user input to str
+        // Copy the user input to command
+        string command;
         for (i = 0; i < len; i++)
-          str[i] = buffer(_promptStartPos + i) & 0x7f;
-        str[len] = '\0';
+          command += buffer(_promptStartPos + i) & 0x7f;
 
         // Add the input to the history
-        addToHistory(str);
+        addToHistory(command.c_str());
 
         // Pass the command to the debugger, and print the result
-        print( instance()->debugger().run(str) + "\n");
-
-        // Get rid of the string buffer
-        delete [] str;
+        print( instance()->debugger().run(command) + "\n");
       }
 
       printPrompt();
@@ -527,27 +521,22 @@ void PromptWidget::specialKeys(int keycode)
       _currentPos = _promptStartPos;
       handled = true;
       break;
-
     case 'd':
       killChar(+1);
       handled = true;
       break;
-
     case 'e':
       _currentPos = _promptEndPos;
       handled = true;
       break;
-
     case 'k':
       killLine(+1);
       handled = true;
       break;
-
     case 'u':
       killLine(-1);
       handled = true;
       break;
-
     case 'w':
       killLastWord();
       handled = true;
@@ -768,7 +757,7 @@ int PromptWidget::printf(const char *format, ...)
 int PromptWidget::vprintf(const char *format, va_list argptr)
 {
   char buf[2048];
-  int count = VSNPRINTF(buf, sizeof(buf), format, argptr);
+  int count = BSPF_vsnprintf(buf, sizeof(buf), format, argptr);
 
   print(buf);
   return count;
@@ -791,10 +780,10 @@ void PromptWidget::putcharIntern(int c)
                       // don't print or advance cursor
                       // there are only 128 TIA colors, but
                       // OverlayColor contains 256 of them
-    textColor = (OverlayColor) ((c & 0x7f) << 1);
+    textColor = (c & 0x7f) << 1;
   }
   else if(c < ' ') { // More colors (the regular GUI ones)
-    textColor = (OverlayColor) (c + 0x100);
+    textColor = c + 0x100;
   }
   else if(c == 0x7f) { // toggle inverse video (DEL char)
     _inverse = !_inverse;
