@@ -8,12 +8,12 @@
 //  SS  SS   tt   ee      ll   ll  aa  aa
 //   SSSS     ttt  eeeee llll llll  aaaaa
 //
-// Copyright (c) 1995-2005 by Bradford W. Mott and the Stella team
+// Copyright (c) 1995-2007 by Bradford W. Mott and the Stella team
 //
 // See the file "license" for information on usage and redistribution of
 // this file, and for a DISCLAIMER OF ALL WARRANTIES.
 //
-// $Id: FrameBuffer.cxx,v 1.77 2006/01/20 13:45:45 stephena Exp $
+// $Id: FrameBuffer.cxx,v 1.116 2007/01/06 16:28:38 stephena Exp $
 //============================================================================
 
 #include <sstream>
@@ -32,35 +32,20 @@
 #include "Launcher.hxx"
 #include "OSystem.hxx"
 
-#ifdef DEVELOPER_SUPPORT
+#ifdef DEBUGGER_SUPPORT
   #include "Debugger.hxx"
 #endif
 
-#if defined(OS2) // FIXME - make proper OS/2 port
-  #define INCL_WIN
-  #include <os2emx.h>
-#endif
-
-#include "stella.xpm"   // The Stella icon
-
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 FrameBuffer::FrameBuffer(OSystem* osystem)
   : myOSystem(osystem),
     myScreen(0),
-    theZoomLevel(2),
-    theMaxZoomLevel(2),
     theAspectRatio(1.0),
     theRedrawTIAIndicator(true),
     myUsePhosphor(false),
     myPhosphorBlend(77),
-    myFrameRate(0),
-    myPauseStatus(false),
-    myMessageTime(0),
-    myMessageText(""),
-    myNumRedraws(0)
+    myInitializedCount(0)
 {
-  myBaseDim.x = myBaseDim.y = myBaseDim.w = myBaseDim.h = 0;
-  myImageDim = myScreenDim = myDesktopDim = myBaseDim;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -74,11 +59,12 @@ void FrameBuffer::initialize(const string& title, uInt32 width, uInt32 height,
 {
   bool isAlreadyInitialized = (SDL_WasInit(SDL_INIT_VIDEO) & SDL_INIT_VIDEO) > 0;
 
+  myBaseDim.x = myBaseDim.y = 0;
   myBaseDim.w = (uInt16) width;
   myBaseDim.h = (uInt16) height;
-  myFrameRate = myOSystem->frameRate();
 
   // Now (re)initialize the SDL video system
+  // These things only have to be done one per FrameBuffer creation
   if(!isAlreadyInitialized)
   {
     Uint32 initflags = SDL_INIT_VIDEO | SDL_INIT_TIMER;
@@ -86,65 +72,49 @@ void FrameBuffer::initialize(const string& title, uInt32 width, uInt32 height,
     if(SDL_Init(initflags) < 0)
       return;
   }
-  setWindowIcon();
-  cls();
+  myInitializedCount++;
 
   // Query the desktop size
   // This is really the job of SDL
   int dwidth = 0, dheight = 0;
-#if defined(OS2)  // FIXME - make proper OS/2 port
-  myDesktopDim.w = WinQuerySysValue(HWND_DESKTOP, SV_CXSCREEN);
-  myDesktopDim.h = WinQuerySysValue(HWND_DESKTOP, SV_CYSCREEN);
-#else
   myOSystem->getScreenDimensions(dwidth, dheight);
   myDesktopDim.w = dwidth;  myDesktopDim.h = dheight;
-#endif
+
+  // Get the aspect ratio for the display if it's been enabled
+  theAspectRatio = 1.0;
+  if(useAspect) setAspectRatio();
 
   // Set fullscreen flag
   mySDLFlags = myOSystem->settings().getBool("fullscreen") ? SDL_FULLSCREEN : 0;
 
-  // Set window title
-  setWindowTitle(title);
-
-  // Get the aspect ratio for the display if it's been enabled
-  theAspectRatio = 1.0;
-  if(useAspect)
-    setAspectRatio();
-
-  // Get the maximum size of a window for the current desktop
-  theMaxZoomLevel = maxWindowSizeForScreen();
-
-  // Check to see if window size will fit in the screen
-  if((uInt32)myOSystem->settings().getInt("zoom") > theMaxZoomLevel)
-    theZoomLevel = theMaxZoomLevel;
-  else
-    theZoomLevel = myOSystem->settings().getInt("zoom");
+  // Set the available scalers for this framebuffer, based on current eventhandler
+  // state and the maximum size of a window for the current desktop
+  setAvailableScalers();
 
   // Initialize video subsystem
+  Scaler scaler;
+  getScaler(scaler, 0, currentScalerName());
+  setScaler(scaler);
   initSubsystem();
 
-  // Set palette for GUI
-  for(int i = 0; i < kNumColors-256; i++)
-    myPalette[i+256] = mapRGB(ourGUIColors[i][0], ourGUIColors[i][1], ourGUIColors[i][2]);
+  // Set window title and icon
+  setWindowTitle(title);
+  setWindowIcon();
 
-  // Set emulation palette if a console exists
-  // Used when entering/exiting debugger
-#ifdef DEVELOPER_SUPPORT
-  if(&myOSystem->console())
-  {
-    enablePhosphor(myOSystem->console().properties().get("Display.Phosphor", true) == "YES");
-    setPalette(myOSystem->console().mediaSource().palette());
-  }
-#endif
+  // And refresh the display
+  myOSystem->eventHandler().refreshDisplay();
 
   // Enable unicode so we can see translated key events
   // (lowercase vs. uppercase characters)
   SDL_EnableUNICODE(1);
 
   // Erase any messages from a previous run
-  myMessageTime = 0;
+  myMessage.counter = 0;
 
-  myUseDirtyRects = myOSystem->settings().getBool("dirtyrects");
+  // Finally, show some information about the framebuffer,
+  // but only on the first initialization
+  if(myInitializedCount == 1 && myOSystem->settings().getBool("showinfo"))
+    cout << about() << endl;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -160,24 +130,14 @@ void FrameBuffer::update()
   {
     case EventHandler::S_EMULATE:
     {
-      bool mediaSourceChanged = false;
+      // Run the console for one frame
+      myOSystem->console().mediaSource().update();
+      if(myOSystem->eventHandler().frying())
+        myOSystem->console().fry();
 
-      // Draw changes to the mediasource
-      if(!myPauseStatus)
-      {
-        myOSystem->console().mediaSource().update();
-        if(myOSystem->eventHandler().frying())
-          myOSystem->console().fry();
-          mediaSourceChanged = true;  // mediasource changed, so force an update
-      }
+      // And update the screen
+      drawMediaSource();
 
-      // Only update the screen if it's been invalidated
-      if(mediaSourceChanged || theRedrawTIAIndicator)
-        drawMediaSource();
-
-      // Draw any pending messages
-      if(myMessageTime > 0 && !myPauseStatus)
-        drawMessage();
       break;  // S_EMULATE
     }
 
@@ -198,10 +158,6 @@ void FrameBuffer::update()
         drawMediaSource();
 
       myOSystem->commandMenu().draw();
-
-      // Draw any pending messages
-      if(myMessageTime > 0 && !myPauseStatus)
-        drawMessage();
       break;  // S_CMDMENU
     }
 
@@ -211,7 +167,7 @@ void FrameBuffer::update()
       break;  // S_LAUNCHER
     }
 
-#ifdef DEVELOPER_SUPPORT
+#ifdef DEBUGGER_SUPPORT
     case EventHandler::S_DEBUGGER:
     {
       myOSystem->debugger().draw();
@@ -224,6 +180,10 @@ void FrameBuffer::update()
       break;
   }
 
+  // Draw any pending messages
+  if(myMessage.counter > 0)
+    drawMessage();
+
   // Do any post-frame stuff
   postFrameUpdate();
 
@@ -232,86 +192,160 @@ void FrameBuffer::update()
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void FrameBuffer::refresh(bool now)
-{
-//  cerr << "refreshTIA() " << myNumRedraws++ << endl;
-  theRedrawTIAIndicator = true;
-  if(now)
-  {
-    myMessageTime = 0;
-    update();
-  }
-}
-
-// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void FrameBuffer::showMessage(const string& message)
+void FrameBuffer::showMessage(const string& message, MessagePosition position,
+                              int color)
 {
   // Erase old messages on the screen
-  if(myMessageTime > 0)
+  if(myMessage.counter > 0)
   {
     theRedrawTIAIndicator = true;
-    drawMediaSource();
+    myOSystem->eventHandler().refreshDisplay();
   }
 
-  myMessageText = message;
-  myMessageTime = myFrameRate << 1;   // Show message for 2 seconds
+  // Precompute the message coordinates
+  myMessage.text    = message;
+  myMessage.counter = myOSystem->frameRate() << 1; // Show message for 2 seconds
+  myMessage.color   = color;
+
+  myMessage.w = myOSystem->font().getStringWidth(myMessage.text) + 10;
+  myMessage.h = myOSystem->font().getFontHeight() + 8;
+
+  switch(position)
+  {
+    case kTopLeft:
+      myMessage.x = 5;
+      myMessage.y = 5;
+      break;
+
+    case kTopCenter:
+      myMessage.x = (myBaseDim.w >> 1) - (myMessage.w >> 1);
+      myMessage.y = 5;
+      break;
+
+    case kTopRight:
+      myMessage.x = myBaseDim.w - myMessage.w - 5;
+      myMessage.y = 5;
+      break;
+
+    case kMiddleLeft:
+      myMessage.x = 5;
+      myMessage.y = (myBaseDim.h >> 1) - (myMessage.h >> 1);
+      break;
+
+    case kMiddleCenter:
+      myMessage.x = (myBaseDim.w >> 1) - (myMessage.w >> 1);
+      myMessage.y = (myBaseDim.h >> 1) - (myMessage.h >> 1);
+      break;
+
+    case kMiddleRight:
+      myMessage.x = myBaseDim.w - myMessage.w - 5;
+      myMessage.y = (myBaseDim.h >> 1) - (myMessage.h >> 1);
+      break;
+
+    case kBottomLeft:
+      myMessage.x = 5;//(myMessage.w >> 1);
+      myMessage.y = myBaseDim.h - myMessage.h - 5;
+      break;
+
+    case kBottomCenter:
+      myMessage.x = (myBaseDim.w >> 1) - (myMessage.w >> 1);
+      myMessage.y = myBaseDim.h - myMessage.h - 5;
+      break;
+
+    case kBottomRight:
+      myMessage.x = myBaseDim.w - myMessage.w - 5;
+      myMessage.y = myBaseDim.h - myMessage.h - 5;
+      break;
+  }
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void FrameBuffer::hideMessage()
 {
   // Erase old messages on the screen
-  if(myMessageTime > 0)
-    refresh(true);
+  if(myMessage.counter > 0)
+  {
+    myMessage.counter = 0;
+    myOSystem->eventHandler().refreshDisplay(true);  // Do this twice for
+    myOSystem->eventHandler().refreshDisplay(true);  // double-buffered modes
+  }
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void FrameBuffer::pause(bool status)
+inline void FrameBuffer::drawMessage()
 {
-  myPauseStatus = status;
+  // Draw the bounded box and text
+  fillRect(myMessage.x+1, myMessage.y+2, myMessage.w-2, myMessage.h-4, kBGColor);
+  box(myMessage.x, myMessage.y+1, myMessage.w, myMessage.h-2, kColor, kColor);
+  drawString(&myOSystem->font(), myMessage.text, myMessage.x+1, myMessage.y+4,
+             myMessage.w, myMessage.color, kTextAlignCenter);
+  myMessage.counter--;
+
+  // Either erase the entire message (when time is reached),
+  // or show again this frame
+  if(myMessage.counter == 0)  // Force an immediate update
+    myOSystem->eventHandler().refreshDisplay(true);
+  else
+    addDirtyRect(myMessage.x, myMessage.y, myMessage.w, myMessage.h);
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void FrameBuffer::setPalette(const uInt32* palette)
+void FrameBuffer::refresh()
+{
+  theRedrawTIAIndicator = true;
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+void FrameBuffer::setTIAPalette(const uInt32* palette)
 {
   int i, j;
 
   // Set palette for normal fill
   for(i = 0; i < 256; ++i)
   {
-    Uint8 r = (Uint8) ((palette[i] & 0x00ff0000) >> 16);
-    Uint8 g = (Uint8) ((palette[i] & 0x0000ff00) >> 8);
-    Uint8 b = (Uint8) (palette[i] & 0x000000ff);
+    Uint8 r = (palette[i] >> 16) & 0xff;
+    Uint8 g = (palette[i] >> 8) & 0xff;
+    Uint8 b = palette[i] & 0xff;
 
-    myPalette[i] = mapRGB(r, g, b);
+    myDefPalette[i] = mapRGB(r, g, b);
   }
 
   // Set palette for phosphor effect
-  if(myUsePhosphor)
+  for(i = 0; i < 256; ++i)
   {
-    for(i = 0; i < 256; ++i)
+    for(j = 0; j < 256; ++j)
     {
-      for(j = 0; j < 256; ++j)
-      {
-        uInt8 ri = (uInt8) ((palette[i] & 0x00ff0000) >> 16);
-        uInt8 gi = (uInt8) ((palette[i] & 0x0000ff00) >> 8);
-        uInt8 bi = (uInt8) (palette[i] & 0x000000ff);
-        uInt8 rj = (uInt8) ((palette[j] & 0x00ff0000) >> 16);
-        uInt8 gj = (uInt8) ((palette[j] & 0x0000ff00) >> 8);
-        uInt8 bj = (uInt8) (palette[j] & 0x000000ff);
+      uInt8 ri = (palette[i] >> 16) & 0xff;
+      uInt8 gi = (palette[i] >> 8) & 0xff;
+      uInt8 bi = palette[i] & 0xff;
+      uInt8 rj = (palette[j] >> 16) & 0xff;
+      uInt8 gj = (palette[j] >> 8) & 0xff;
+      uInt8 bj = palette[j] & 0xff;
 
-        Uint8 r = (Uint8) getPhosphor(ri, rj);
-        Uint8 g = (Uint8) getPhosphor(gi, gj);
-        Uint8 b = (Uint8) getPhosphor(bi, bj);
+      Uint8 r = (Uint8) getPhosphor(ri, rj);
+      Uint8 g = (Uint8) getPhosphor(gi, gj);
+      Uint8 b = (Uint8) getPhosphor(bi, bj);
 
-        myAvgPalette[i][j] = mapRGB(r, g, b);
-      }
+      myAvgPalette[i][j] = mapRGB(r, g, b);
     }
   }
 
   theRedrawTIAIndicator = true;
 }
 
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+void FrameBuffer::setUIPalette(const uInt32* palette)
+{
+  // Set palette for GUI
+  for(int i = 0; i < kNumColors-256; ++i)
+  {
+    Uint8 r = (palette[i] >> 16) & 0xff;
+    Uint8 g = (palette[i] >> 8) & 0xff;
+    Uint8 b = palette[i] & 0xff;
+    myDefPalette[i+256] = mapRGB(r, g, b);
+  }
+}
+
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void FrameBuffer::toggleFullscreen()
 {
@@ -332,48 +366,38 @@ void FrameBuffer::setFullscreen(bool enable)
   if(!createScreen())
     return;
 
+  myOSystem->eventHandler().refreshDisplay();
   setCursorState();
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void FrameBuffer::resize(Size size, Int8 zoom)
+bool FrameBuffer::scale(int direction, const string& type)
 {
-  switch(size)
+  Scaler newScaler;
+  const string& currentScaler = (direction == 0 ? type : currentScalerName());
+  getScaler(newScaler, direction, currentScaler);
+
+  // Only update the scaler if it's changed from the old one
+  if(currentScaler != string(newScaler.comparitor))
   {
-    case PreviousSize:   // decrease size
-      if(myOSystem->settings().getBool("fullscreen"))
-        return;
-      if(theZoomLevel == 1)
-        theZoomLevel = theMaxZoomLevel;
-      else
-        theZoomLevel--;
-      break;
+    setScaler(newScaler);
+    if(!createScreen())
+      return false;
 
-    case NextSize:       // increase size
-      if(myOSystem->settings().getBool("fullscreen"))
-        return;
-      if(theZoomLevel == theMaxZoomLevel)
-        theZoomLevel = 1;
-      else
-        theZoomLevel++;
-      break;
+    EventHandler::State state = myOSystem->eventHandler().state();
+    bool inTIAMode = (state == EventHandler::S_EMULATE ||
+                      state == EventHandler::S_MENU    ||
+                      state == EventHandler::S_CMDMENU);
 
-    case GivenSize:      // use 'zoom' quantity
-      if(zoom < 1)
-        theZoomLevel = 1;
-      else if((uInt32)zoom > theMaxZoomLevel)
-        theZoomLevel = theMaxZoomLevel;
-      else
-        theZoomLevel = zoom;
-      break;
+    myOSystem->eventHandler().refreshDisplay();
+    showMessage(newScaler.name);
 
-    default:   // should never happen
-      return;
-      break;
+    if(inTIAMode)
+      myOSystem->settings().setString("scale_tia", newScaler.comparitor);
+    else
+      myOSystem->settings().setString("scale_ui", newScaler.comparitor);
   }
-
-  if(!createScreen())
-    return;
+  return true;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -391,7 +415,6 @@ void FrameBuffer::setCursorState()
     case EventHandler::S_EMULATE:
       showCursor(false);
       break;
-
     default:
       showCursor(true);
   }
@@ -421,36 +444,6 @@ bool FrameBuffer::fullScreen()
   return myOSystem->settings().getBool("fullscreen");
 }
 
-// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-uInt32 FrameBuffer::maxWindowSizeForScreen()
-{
-  uInt32 sWidth     = myDesktopDim.w;
-  uInt32 sHeight    = myDesktopDim.h;
-  uInt32 multiplier = 10;
-
-  // If screenwidth or height could not be found, use default zoom value
-  if(sWidth == 0 || sHeight == 0)
-    return 4;
-
-  bool found = false;
-  while(!found && (multiplier > 0))
-  {
-    // Figure out the desired size of the window
-    uInt32 width  = (uInt32) (myBaseDim.w * multiplier * theAspectRatio);
-    uInt32 height = myBaseDim.h * multiplier;
-
-    if((width < sWidth) && (height < sHeight))
-      found = true;
-    else
-      multiplier--;
-  }
-
-  if(found)
-    return multiplier;
-  else
-    return 1;
-}
-
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void FrameBuffer::setWindowTitle(const string& title)
 {
@@ -461,6 +454,8 @@ void FrameBuffer::setWindowTitle(const string& title)
 void FrameBuffer::setWindowIcon()
 {
 #ifndef MAC_OSX
+  #include "stella.xpm"   // The Stella icon
+
   // Set the window icon
   uInt32 w, h, ncols, nbytes;
   uInt32 rgba[256], icon[32 * 32];
@@ -516,34 +511,9 @@ void FrameBuffer::setWindowIcon()
 #endif
 }
 
-// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-inline void FrameBuffer::drawMessage()
-{
-  int w = myOSystem->font().getStringWidth(myMessageText) + 10;
-  int h = myOSystem->font().getFontHeight() + 8;
-  int x = (myBaseDim.w >> 1) - (w >> 1);
-  int y = myBaseDim.h - h - 10/2;
-
-  // Draw the bounded box and text
-  fillRect(x+1, y+2, w-2, h-4, kBGColor);
-  box(x, y+1, w, h-2, kColor, kColor);
-  drawString(&myOSystem->font(), myMessageText, x+1, y+4, w, kTextColor, kTextAlignCenter);
-  myMessageTime--;
-
-  // Either erase the entire message (when time is reached),
-  // or show again this frame
-  if(myMessageTime == 0)
-  {
-	theRedrawTIAIndicator = true;
-	drawMediaSource();
-  }
-  else
-	addDirtyRect(x, y, w, h);
-}
-
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void FrameBuffer::box(uInt32 x, uInt32 y, uInt32 w, uInt32 h,
-                      OverlayColor colorA, OverlayColor colorB)
+                      int colorA, int colorB)
 {
   hLine(x + 1, y,     x + w - 2, colorA);
   hLine(x,     y + 1, x + w - 1, colorA);
@@ -558,7 +528,7 @@ void FrameBuffer::box(uInt32 x, uInt32 y, uInt32 w, uInt32 h,
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void FrameBuffer::frameRect(uInt32 x, uInt32 y, uInt32 w, uInt32 h,
-                            OverlayColor color, FrameStyle style)
+                            int color, FrameStyle style)
 {
   switch(style)
   {
@@ -570,7 +540,7 @@ void FrameBuffer::frameRect(uInt32 x, uInt32 y, uInt32 w, uInt32 h,
       break;
 
     case kDashLine:
-      unsigned int i, skip, lwidth = 1;
+      unsigned int i, skip, lwidth = 0;
 
       for(i = x, skip = 1; i < x+w-1; i=i+lwidth+1, ++skip)
       {
@@ -595,7 +565,7 @@ void FrameBuffer::frameRect(uInt32 x, uInt32 y, uInt32 w, uInt32 h,
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void FrameBuffer::drawString(const GUI::Font* font, const string& s,
                              int x, int y, int w,
-                             OverlayColor color, TextAlignment align,
+                             int color, TextAlignment align,
                              int deltax, bool useEllipsis)
 {
   const int leftX = x, rightX = x + w;
@@ -678,12 +648,107 @@ uInt8 FrameBuffer::getPhosphor(uInt8 c1, uInt8 c2)
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-const uInt8 FrameBuffer::ourGUIColors[kNumColors-256][3] = {
-  { 104, 104, 104 },  // kColor
-  {   0,   0,   0 },  // kBGColor
-  {  64,  64,  64 },  // kShadowColor
-  { 200, 200, 255 },  // kHiliteColor
-  {  32, 160,  32 },  // kTextColor
-  {   0, 255,   0 },  // kTextColorHi
-  { 200,   0,   0 }   // kTextColorEm
+int FrameBuffer::maxWindowSizeForScreen()
+{
+  uInt32 sWidth     = myDesktopDim.w;
+  uInt32 sHeight    = myDesktopDim.h;
+  uInt32 multiplier = 10;
+
+  // If screenwidth or height could not be found, use default zoom value
+  if(sWidth == 0 || sHeight == 0)
+    return 4;
+
+  bool found = false;
+  while(!found && (multiplier > 0))
+  {
+    // Figure out the desired size of the window
+    uInt32 width  = (uInt32) (myBaseDim.w * multiplier * theAspectRatio);
+    uInt32 height = myBaseDim.h * multiplier;
+
+    if((width < sWidth) && (height < sHeight))
+      found = true;
+    else
+      multiplier--;
+  }
+
+  if(found)
+    return multiplier;
+  else
+    return 1;
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+void FrameBuffer::setAvailableScalers()
+{
+  /** Different emulation modes support different scalers, and the size
+      of the current desktop also determines how much a window can be
+      zoomed. */
+  int maxsize = maxWindowSizeForScreen();
+  myScalerList.clear();
+
+  for(int i = 0; i < kScalerListSize; ++i)
+    if(ourScalers[i].zoom <= maxsize)
+      myScalerList.push_back(&ourScalers[i]);
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+void FrameBuffer::getScaler(Scaler& scaler, int direction, const string& name)
+{
+  // First search for the scaler specified by name
+  int pos = -1;
+  for(unsigned int i = 0; i < myScalerList.size(); ++i)
+  {
+    if(BSPF_strcasecmp(myScalerList[i]->name, name.c_str()) == 0)
+    {
+      pos = i;
+      break;
+    }
+  }
+
+  // If we found a scaler, look at direction
+  if(pos >= 0)
+  {
+    switch(direction)
+    {
+      case 0:   // actual scaler specified in 'name'
+        // pos is already set from above
+        break;
+      case -1:  // previous scaler from one specified in 'name'
+        pos--;
+        if(pos < 0) pos = myScalerList.size() - 1;
+        break;
+      case +1:  // next scaler from one specified in 'name'
+        pos = (pos + 1) % myScalerList.size();
+        break;
+    }
+    scaler = *(myScalerList[pos]);
+  }
+  else
+  {
+    // Otherwise, get the largest scaler that's available
+    scaler = *(myScalerList[myScalerList.size()-1]);
+  }
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+const string& FrameBuffer::currentScalerName()
+{
+  EventHandler::State state = myOSystem->eventHandler().state();
+  bool inTIAMode = (state == EventHandler::S_EMULATE ||
+                    state == EventHandler::S_MENU    ||
+                    state == EventHandler::S_CMDMENU);
+
+  return (inTIAMode ?
+     myOSystem->settings().getString("scale_tia") :
+     myOSystem->settings().getString("scale_ui") );
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Scaler FrameBuffer::ourScalers[kScalerListSize] = {
+  { kZOOM1X,  "Zoom1x", "zoom1x", 1 },
+  { kZOOM2X,  "Zoom2x", "zoom2x", 2 },
+  { kZOOM3X,  "Zoom3x", "zoom3x", 3 },
+  { kZOOM4X,  "Zoom4x", "zoom4x", 4 },
+  { kZOOM5X,  "Zoom5x", "zoom5x", 5 },
+  { kZOOM6X,  "Zoom6x", "zoom6x", 6 }
 };
