@@ -8,12 +8,12 @@
 //  SS  SS   tt   ee      ll   ll  aa  aa
 //   SSSS     ttt  eeeee llll llll  aaaaa
 //
-// Copyright (c) 1995-2005 by Bradford W. Mott and the Stella team
+// Copyright (c) 1995-2007 by Bradford W. Mott and the Stella team
 //
 // See the file "license" for information on usage and redistribution of
 // this file, and for a DISCLAIMER OF ALL WARRANTIES.
 //
-// $Id: Settings.cxx,v 1.76 2006/01/15 20:46:19 stephena Exp $
+// $Id: Settings.cxx,v 1.115 2007/01/03 17:37:46 stephena Exp $
 //============================================================================
 
 #include <cassert>
@@ -27,69 +27,69 @@
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Settings::Settings(OSystem* osystem)
-    : myOSystem(osystem)
+  : myOSystem(osystem)
 {
   // Add this settings object to the OSystem
   myOSystem->attach(this);
 
-  // First create the settings array
-  myCapacity = 30;
-  mySettings = new Setting[myCapacity];
-  mySize = 0;
+  // Add options that are common to all versions of Stella
+  setInternal("video", "soft");
+  setInternal("dirtyrects", "false"); // This only becomes useful at high resolutions
 
-  // Now fill it with options that are common to all versions of Stella
-  set("video", "soft");
-  set("dirtyrects", "true");
-  set("ppblend", "77");
+  setInternal("gl_filter", "nearest");
+  setInternal("gl_aspect", "2.0");
+  setInternal("gl_fsmax", "true");
+  setInternal("gl_lib", "libGL.so");
+  setInternal("gl_vsync", "false");
 
-  set("gl_filter", "nearest");
-  set("gl_aspect", "2.0");
-  set("gl_fsmax", "false");
-  set("gl_lib", "");
+  setInternal("scale_ui", "zoom2x");
+  setInternal("scale_tia", "zoom2x");
+  setInternal("fullscreen", "false");
+  setInternal("center", "true");
+  setInternal("grabmouse", "false");
+  setInternal("palette", "standard");
+  setInternal("colorloss", "false");
 
-  set("zoom", "2");
-  set("fullscreen", "false");
-  set("center", "true");
-  set("grabmouse", "false");
-  set("palette", "standard");
-  set("debugheight", "0");
+  setInternal("sound", "true");
+  setInternal("fragsize", "512");
+  setInternal("freq", "31400");
+  setInternal("tiafreq", "31400");
+  setInternal("volume", "100");
+  setInternal("clipvol", "true");
 
-  set("sound", "true");
-  set("fragsize", "512");
-  set("freq", "31400");
-  set("tiafreq", "31400");
-  set("volume", "100");
-  set("clipvol", "true");
+  setInternal("keymap", "");
+  setInternal("joymap", "");
+  setInternal("joyaxismap", "");
+  setInternal("joyhatmap", "");
+  setInternal("paddle", "0");
+  setInternal("sa1", "left");
+  setInternal("sa2", "right");
+  setInternal("p0speed", "50");
+  setInternal("p1speed", "50");
+  setInternal("p2speed", "50");
+  setInternal("p3speed", "50");
+  setInternal("pthresh", "600");
 
-  set("keymap", "");
-  set("joymap", "");
-  set("joyaxismap", "");
-  set("joyhatmap", "");
-  set("paddle", "0");
-  set("sa1", "left");
-  set("sa2", "right");
-  set("joymouse", "false");
-  set("p1speed", "50");
-  set("p2speed", "50");
-  set("p3speed", "50");
-  set("p4speed", "50");
+  setInternal("showinfo", "false");
 
-  set("showinfo", "false");
+  setInternal("ssdir", string(".") + BSPF_PATH_SEPARATOR);
+  setInternal("sssingle", "false");
 
-  set("ssdir", "");
-  set("ssname", "romname");
-  set("sssingle", "false");
-
-  set("romdir", "");
-  set("lastrom", "");
-  set("modtime", "");  // romdir last modification time
+  setInternal("romdir", "");
+  setInternal("rombrowse", "true");
+  setInternal("lastrom", "");
+  setInternal("modtime", "");
+  setInternal("debugheight", "0");
+  setInternal("launchersize", "2");
+  setInternal("uipalette", "0");
+  setInternal("autoslot", "false");
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Settings::~Settings()
 {
-  // Free the settings array
-  delete[] mySettings;
+  myInternalSettings.clear();
+  myExternalSettings.clear();
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -98,12 +98,7 @@ void Settings::loadConfig()
   string line, key, value;
   string::size_type equalPos, garbage;
 
-  // FIXME - add infrastructure to print output
-#ifdef PSP_DEBUG
-  fprintf(stdout,"Settings::loadConfig '%s'\n",myOSystem->configInputFilename().c_str());
-#endif
-
-  ifstream in(myOSystem->configInputFilename().c_str());
+  ifstream in(myOSystem->configFile().c_str());
   if(!in || !in.is_open())
   {
     cout << "Error: Couldn't load settings file\n";
@@ -135,8 +130,8 @@ void Settings::loadConfig()
       continue;
 
     // Only settings which have been previously set are valid
-    if(contains(key))
-      set(key, value);
+    if(int idx = getInternalPos(key) != -1)
+      setInternal(key, value, idx, true);
   }
 
   in.close();
@@ -163,27 +158,32 @@ bool Settings::loadCommandLine(int argc, char** argv)
     }
     else if(key == "listrominfo")
     {
-      set(key, "true", false);  // this confusing line means set 'listrominfo'
-      return true;              // to true, but don't save to the settings file
+      setExternal(key, "true");
+      return true;
+    }
+    else if(key == "rominfo")
+    {
+      setExternal(key, "true");
+      return true;
     }
     else if(key == "debug") // this doesn't make Stella exit
     {
-      set(key, "true", false); // don't save this to the config file either
+      setExternal(key, "true");
       return true;
     }
     else if(key == "holdreset") // this doesn't make Stella exit
     {
-      set(key, "true", false); // don't save this to the config file either
+      setExternal(key, "true");
       return true;
     }
     else if(key == "holdselect") // this doesn't make Stella exit
     {
-      set(key, "true", false); // don't save this to the config file either
+      setExternal(key, "true");
       return true;
     }
     else if(key == "holdbutton0") // this doesn't make Stella exit
     {
-      set(key, "true", false); // don't save this to the config file either
+      setExternal(key, "true");
       return true;
     }
 
@@ -196,7 +196,10 @@ bool Settings::loadCommandLine(int argc, char** argv)
 
     // Settings read from the commandline must not be saved to 
     // the rc-file, unless they were previously set
-    set(key, value, false);
+    if(int idx = getInternalPos(key) != -1)
+      setInternal(key, value, idx);   // don't set initialValue here
+    else
+      setExternal(key, value);
   }
 
   return true;
@@ -210,57 +213,53 @@ void Settings::validate()
 
   s = getString("video");
   if(s != "soft" && s != "gl")
-    set("video", "soft");
+    setInternal("video", "soft");
 
 #ifdef DISPLAY_OPENGL
   s = getString("gl_filter");
   if(s != "linear" && s != "nearest")
-    set("gl_filter", "nearest");
+    setInternal("gl_filter", "nearest");
 
   float f = getFloat("gl_aspect");
   if(f < 1.1 || f > 2.0)
-    set("gl_aspect", "2.0");
+    setInternal("gl_aspect", "2.0");
 #endif
 
 #ifdef SOUND_SUPPORT
-  i = getInt("fragsize");
-  if(i != 256 && i != 512 && i != 1024 && i != 2048 && i != 4096)
-  #ifdef WIN32
-    set("fragsize", "2048");
-  #else
-    set("fragsize", "512");
-  #endif
-
   i = getInt("volume");
   if(i < 0 || i > 100)
-    set("volume", "100");
+    setInternal("volume", "100");
   i = getInt("freq");
   if(i < 0 || i > 48000)
-    set("freq", "31400");
+    setInternal("freq", "31400");
   i = getInt("tiafreq");
   if(i < 0 || i > 48000)
-    set("tiafreq", "31400");
+    setInternal("tiafreq", "31400");
 #endif
 
-  i = getInt("zoom");
-  if(i < 1 || i > 6)
-    set("zoom", "2");
+  s = getString("scale_ui");
+  if(s != "zoom1x" && s != "zoom2x" && s != "zoom3x" &&
+     s != "zoom4x" && s != "zoom5x" && s != "zoom6x")
+    setInternal("scale_ui", "zoom1x");
+
+  s = getString("scale_tia");
+  if(s != "zoom1x" && s != "zoom2x" && s != "zoom3x" &&
+     s != "zoom4x" && s != "zoom5x" && s != "zoom6x")
+    setInternal("scale_tia", "zoom1x");
 
   i = getInt("paddle");
   if(i < 0 || i > 3)
-    set("paddle", "0");
+    setInternal("paddle", "0");
+
+  i = getInt("pthresh");
+  if(i < 400)
+    setInternal("pthresh", "400");
+  else if(i > 800)
+    setInternal("pthresh", "800");
 
   s = getString("palette");
-  if(s != "standard" && s != "original" && s != "z26")
-    set("palette", "standard");
-
-  i = getInt("ppblend");
-  if(i < 0) set("ppblend", "0");
-  if(i > 100) set("ppblend", "100");
-
-  s = getString("romname");
-  if(s != "romname" && s != "md5sum")
-    set("ssname", "romname");
+  if(s != "standard" && s != "original" && s != "z26" && s != "user")
+    setInternal("palette", "standard");
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -280,27 +279,29 @@ void Settings::usage()
   #ifdef DISPLAY_OPENGL
     << "                 gl              SDL OpenGL mode\n"
     << endl
+    << "  -gl_lib       <name>         Specify the OpenGL library\n"
     << "  -gl_filter    <type>         Type is one of the following:\n"
     << "                 nearest         Normal scaling (GL_NEAREST)\n"
     << "                 linear          Blurred scaling (GL_LINEAR)\n"
     << "  -gl_aspect    <number>       Scale the width by the given amount\n"
     << "  -gl_fsmax     <1|0>          Use the largest available screenmode in fullscreen OpenGL\n"
-    << "  -gl_lib       <filename>     Specify the OpenGL library\n"
+    << "  -gl_vsync     <1|0>          Enable synchronize to vertical blank interrupt\n"
     << endl
   #endif
-    << "  -zoom         <size>         Makes window be 'size' times normal\n"
+    << "  -scale_tia    <scaler>       Use the specified scaler in emulation mode\n"
+    << "  -scale_ui     <scaler>       Use the specified scaler in non-emulation mode (ROM browser/debugger)\n"
     << "  -fullscreen   <1|0>          Play the game in fullscreen mode\n"
     << "  -center       <1|0>          Centers game window (if possible)\n"
     << "  -grabmouse    <1|0>          Keeps the mouse in the game window\n"
     << "  -palette      <original|     Use the specified color palette\n"
     << "                 standard|\n"
-    << "                 z26>\n"
+    << "                 z26|\n"
+    << "                 user>\n"
+    << "  -colorloss    <1|0>          Enable PAL color-loss effect\n"
     << "  -framerate    <number>       Display the given number of frames per second\n"
-    << "  -ppblend      <number>       Set blending for phosphor effect, if enabled (0-100)\n"
     << endl
   #ifdef SOUND_SUPPORT
     << "  -sound        <1|0>          Enable sound generation\n"
-    << "  -channels     <1|2>          Enable mono or stereo sound\n"
     << "  -fragsize     <number>       The size of sound fragments (must be a power of two)\n"
     << "  -freq         <number>       Set sound sample output frequency (0 - 48000)\n"
     << "  -tiafreq      <number>       Set sound sample generation frequency (0 - 48000)\n"
@@ -313,23 +314,25 @@ void Settings::usage()
     << "  -paddle       <0|1|2|3>      Indicates which paddle the mouse should emulate\n"
     << "  -sa1          <left|right>   Stelladaptor 1 emulates specified joystick port\n"
     << "  -sa2          <left|right>   Stelladaptor 2 emulates specified joystick port\n"
-    << "  -joymouse     <1|0>          Enable mouse emulation using joystick in GUI\n"
+    << "  -p0speed      <number>       Speed of emulated mouse movement for paddle 0 (0-100)\n"
     << "  -p1speed      <number>       Speed of emulated mouse movement for paddle 1 (0-100)\n"
     << "  -p2speed      <number>       Speed of emulated mouse movement for paddle 2 (0-100)\n"
     << "  -p3speed      <number>       Speed of emulated mouse movement for paddle 3 (0-100)\n"
-    << "  -p4speed      <number>       Speed of emulated mouse movement for paddle 4 (0-100)\n"
+    << "  -pthresh      <number>       Set threshold for eliminating paddle jitter\n"
+    << "  -rombrowse    <1|0>          Use ROM browser mode (shows files and folders)\n"
+    << "  -autoslot     <1|0>          Automatically switch to next save slot when state saving\n"
   #ifdef UNIX
     << "  -accurate     <1|0>          Accurate game timing (uses more CPU)\n"
   #endif
-  #ifdef SNAPSHOT_SUPPORT
     << "  -ssdir        <path>         The directory to save snapshot files to\n"
-    << "  -ssname       <name>         How to name the snapshot (romname or md5sum)\n"
     << "  -sssingle     <1|0>          Generate single snapshot instead of many\n"
     << endl
-  #endif
     << "  -listrominfo                 Display contents of stella.pro, one line per ROM entry\n"
+    << "  -rominfo      <rom>          Display detailed information for the given ROM\n"
+    << "  -launchersize <1|2|3>        Set the size of the ROM launcher\n"
+    << "  -uipalette    <1|2>          Used the specified palette for UI elements\n"
     << "  -help                        Show the text you're now reading\n"
-  #ifdef DEVELOPER_SUPPORT
+  #ifdef DEBUGGER_SUPPORT
     << endl
     << " The following options are meant for developers\n"
     << " Arguments are more fully explained in the manual\n"
@@ -350,12 +353,14 @@ void Settings::usage()
     << "   -lc          <arg>          Sets the 'Controller.Left' property\n"
     << "   -rc          <arg>          Sets the 'Controller.Right' property\n"
     << "   -bc          <arg>          Same as using both -lc and -rc\n"
+    << "   -cp          <arg>          Sets the 'Controller.SwapPaddles' property\n"
     << "   -format      <arg>          Sets the 'Display.Format' property\n"
     << "   -xstart      <arg>          Sets the 'Display.XStart' property\n"
     << "   -ystart      <arg>          Sets the 'Display.YStart' property\n"
     << "   -width       <arg>          Sets the 'Display.Width' property\n"
     << "   -height      <arg>          Sets the 'Display.Height' property\n"
     << "   -pp          <arg>          Sets the 'Display.Phosphor' property\n"
+    << "   -ppblend     <arg>          Sets the 'Display.PPBlend' property\n"
     << "   -hmove       <arg>          Sets the 'Emulation.HmoveBlanks' property\n"
   #endif
     << endl;
@@ -365,7 +370,22 @@ void Settings::usage()
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void Settings::saveConfig()
 {
-  ofstream out(myOSystem->configOutputFilename().c_str());
+  // Do a quick scan of the internal settings to see if any have
+  // changed.  If not, we don't need to save them at all.
+  bool settingsChanged = false;
+  for(unsigned int i = 0; i < myInternalSettings.size(); ++i)
+  {
+    if(myInternalSettings[i].value != myInternalSettings[i].initialValue)
+    {
+      settingsChanged = true;
+      break;
+    }
+  }
+
+  if(!settingsChanged)
+    return;
+
+  ofstream out(myOSystem->configFile().c_str());
   if(!out || !out.is_open())
   {
     cout << "Error: Couldn't save settings file\n";
@@ -388,153 +408,254 @@ void Settings::saveConfig()
       << ";" << endl;
 
   // Write out each of the key and value pairs
-  for(unsigned int i = 0; i < mySize; ++i)
-    if(mySettings[i].save)
-      out << mySettings[i].key << " = " << mySettings[i].value << endl;
+  for(unsigned int i = 0; i < myInternalSettings.size(); ++i)
+  {
+    out << myInternalSettings[i].key << " = " <<
+           myInternalSettings[i].value << endl;
+  }
 
   out.close();
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void Settings::set(const string& key, const string& value, bool save)
-{
-  // See if the setting already exists
-  for(unsigned int i = 0; i < mySize; ++i)
-  {
-    // If a key is already present in the array, then we assume
-    // that it was set by the emulation core and must be saved
-    // to the rc-file.
-    if(key == mySettings[i].key)
-    {
-      mySettings[i].value = value;
-      mySettings[i].save  = true;
-      return;
-    }
-  }
-
-  // See if the array needs to be resized
-  if(mySize == myCapacity)
-  {
-    // Yes, so we'll make the array twice as large
-    Setting* newSettings = new Setting[myCapacity * 2];
-
-    for(unsigned int i = 0; i < mySize; ++i)
-    {
-      newSettings[i] = mySettings[i];
-    }
-
-    delete[] mySettings;
-
-    mySettings = newSettings;
-    myCapacity *= 2;
-  } 
-
-  // Add new property to the array
-  mySettings[mySize].key   = key;
-  mySettings[mySize].value = value;
-  mySettings[mySize].save  = save;
-
-  ++mySize;
-}
-
-// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void Settings::setInt(const string& key, const int value, bool save)
+void Settings::setInt(const string& key, const int value)
 {
   ostringstream stream;
   stream << value;
 
-  set(key, stream.str(), save);
+  if(int idx = getInternalPos(key) != -1)
+    setInternal(key, stream.str(), idx);
+  else
+    setExternal(key, stream.str());
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void Settings::setFloat(const string& key, const float value, bool save)
+void Settings::setFloat(const string& key, const float value)
 {
   ostringstream stream;
   stream << value;
 
-  set(key, stream.str(), save);
+  if(int idx = getInternalPos(key) != -1)
+    setInternal(key, stream.str(), idx);
+  else
+    setExternal(key, stream.str());
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void Settings::setBool(const string& key, const bool value, bool save)
+void Settings::setBool(const string& key, const bool value)
 {
   ostringstream stream;
   stream << value;
 
-  set(key, stream.str(), save);
+  if(int idx = getInternalPos(key) != -1)
+    setInternal(key, stream.str(), idx);
+  else
+    setExternal(key, stream.str());
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void Settings::setString(const string& key, const string& value, bool save)
+void Settings::setString(const string& key, const string& value)
 {
-  ostringstream stream;
-  stream << value;
-
-  set(key, stream.str(), save);
+  if(int idx = getInternalPos(key) != -1)
+    setInternal(key, value, idx);
+  else
+    setExternal(key, value);
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 int Settings::getInt(const string& key) const
 {
   // Try to find the named setting and answer its value
-  for(unsigned int i = 0; i < mySize; ++i)
-    if(key == mySettings[i].key)
-      return (int) atoi(mySettings[i].value.c_str());
-
-  return 0;
+  int idx = -1;
+  if((idx = getInternalPos(key)) != -1)
+    return (int) atoi(myInternalSettings[idx].value.c_str());
+  else if((idx = getExternalPos(key)) != -1)
+    return (int) atoi(myExternalSettings[idx].value.c_str());
+  else
+    return -1;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 float Settings::getFloat(const string& key) const
 {
   // Try to find the named setting and answer its value
-  for(unsigned int i = 0; i < mySize; ++i)
-    if(key == mySettings[i].key)
-      return (float) atof(mySettings[i].value.c_str());
-
-  return -1.0;
+  int idx = -1;
+  if((idx = getInternalPos(key)) != -1)
+    return (float) atof(myInternalSettings[idx].value.c_str());
+  else if((idx = getExternalPos(key)) != -1)
+    return (float) atof(myExternalSettings[idx].value.c_str());
+  else
+    return -1.0;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 bool Settings::getBool(const string& key) const
 {
   // Try to find the named setting and answer its value
-  for(unsigned int i = 0; i < mySize; ++i)
+  int idx = -1;
+  if((idx = getInternalPos(key)) != -1)
   {
-    if(key == mySettings[i].key)
-    {
-      if(mySettings[i].value == "1" || mySettings[i].value == "true")
-        return true;
-      else if(mySettings[i].value == "0" || mySettings[i].value == "false")
-        return false;
-      else
-        return false;
-    }
+    const string& value = myInternalSettings[idx].value;
+    if(value == "1" || value == "true")
+      return true;
+    else if(value == "0" || value == "false")
+      return false;
+    else
+      return false;
   }
-
-  return false;
+  else if((idx = getExternalPos(key)) != -1)
+  {
+    const string& value = myExternalSettings[idx].value;
+    if(value == "1" || value == "true")
+      return true;
+    else if(value == "0" || value == "false")
+      return false;
+    else
+      return false;
+  }
+  else
+    return false;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 const string& Settings::getString(const string& key) const
 {
   // Try to find the named setting and answer its value
-  for(unsigned int i = 0; i < mySize; ++i)
-    if(key == mySettings[i].key)
-      return mySettings[i].value;
-
-  return EmptyString;
+  int idx = -1;
+  if((idx = getInternalPos(key)) != -1)
+    return myInternalSettings[idx].value;
+  else if((idx = getExternalPos(key)) != -1)
+    return myExternalSettings[idx].value;
+  else
+    return EmptyString;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-bool Settings::contains(const string& key)
+int Settings::getInternalPos(const string& key) const
 {
-  // Try to find the named setting
-  for(unsigned int i = 0; i < mySize; ++i)
-    if(key == mySettings[i].key)
-      return true;
+  for(unsigned int i = 0; i < myInternalSettings.size(); ++i)
+    if(myInternalSettings[i].key == key)
+      return i;
 
-  return false;
+  return -1;
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+int Settings::getExternalPos(const string& key) const
+{
+  for(unsigned int i = 0; i < myExternalSettings.size(); ++i)
+    if(myExternalSettings[i].key == key)
+      return i;
+
+  return -1;
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+int Settings::setInternal(const string& key, const string& value,
+                          int pos, bool useAsInitial)
+{
+  int idx = -1;
+
+  if(pos != -1 && pos >= 0 && pos < (int)myInternalSettings.size() &&
+     myInternalSettings[pos].key == key)
+  {
+    idx = pos;
+  }
+  else
+  {
+    for(unsigned int i = 0; i < myInternalSettings.size(); ++i)
+    {
+      if(myInternalSettings[i].key == key)
+      {
+        idx = i;
+        break;
+      }
+    }
+  }
+
+  if(idx != -1)
+  {
+    myInternalSettings[idx].key   = key;
+    myInternalSettings[idx].value = value;
+    if(useAsInitial) myInternalSettings[idx].initialValue = value;
+
+    /*cerr << "modify internal: key = " << key
+         << ", value  = " << value
+         << ", ivalue = " << myInternalSettings[idx].initialValue
+         << " @ index = " << idx
+         << endl;*/
+  }
+  else
+  {
+    Setting setting;
+    setting.key   = key;
+    setting.value = value;
+    if(useAsInitial) setting.initialValue = value;
+
+    myInternalSettings.push_back(setting);
+    idx = myInternalSettings.size() - 1;
+
+    /*cerr << "insert internal: key = " << key
+         << ", value  = " << value
+         << ", ivalue = " << setting.initialValue
+         << " @ index = " << idx
+         << endl;*/
+  }
+
+  return idx;
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+int Settings::setExternal(const string& key, const string& value,
+                          int pos, bool useAsInitial)
+{
+  int idx = -1;
+
+  if(pos != -1 && pos >= 0 && pos < (int)myExternalSettings.size() &&
+     myExternalSettings[pos].key == key)
+  {
+    idx = pos;
+  }
+  else
+  {
+    for(unsigned int i = 0; i < myExternalSettings.size(); ++i)
+    {
+      if(myExternalSettings[i].key == key)
+      {
+        idx = i;
+        break;
+      }
+    }
+  }
+
+  if(idx != -1)
+  {
+    myExternalSettings[idx].key   = key;
+    myExternalSettings[idx].value = value;
+    if(useAsInitial) myExternalSettings[idx].initialValue = value;
+
+    /*cerr << "modify external: key = " << key
+         << ", value = " << value
+         << " @ index = " << idx
+         << endl;*/
+  }
+  else
+  {
+    Setting setting;
+    setting.key   = key;
+    setting.value = value;
+    if(useAsInitial) setting.initialValue = value;
+
+    myExternalSettings.push_back(setting);
+    idx = myExternalSettings.size() - 1;
+
+    /*cerr << "insert external: key = " << key
+         << ", value = " << value
+         << " @ index = " << idx
+         << endl;*/
+  }
+
+  return idx;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
