@@ -8,12 +8,12 @@
 //  SS  SS   tt   ee      ll   ll  aa  aa
 //   SSSS     ttt  eeeee llll llll  aaaaa
 //
-// Copyright (c) 1995-2005 by Bradford W. Mott and the Stella team
+// Copyright (c) 1995-2007 by Bradford W. Mott and the Stella team
 //
 // See the file "license" for information on usage and redistribution of
 // this file, and for a DISCLAIMER OF ALL WARRANTIES.
 //
-// $Id: SettingsUNIX.cxx,v 1.16 2006/01/14 21:36:29 stephena Exp $
+// $Id: SettingsUNIX.cxx,v 1.21 2007/01/01 18:04:55 stephena Exp $
 //============================================================================
 
 #include "bspf.hxx"
@@ -25,8 +25,12 @@ SettingsUNIX::SettingsUNIX(OSystem* osystem)
   : Settings(osystem)
 {
   // This argument is only valid for Linux/UNIX, and will eventually be removed
-  set("accurate", "false");
-  set("gl_lib", "libGL.so");
+  setInternal("accurate", "false");
+  setInternal("gl_lib", "libGL.so");
+  // Most Linux GL implementations don't support this yet
+  setInternal("gl_vsync", "false");
+  // For whatever reason, this is very efficient in Linux
+  setInternal("dirtyrects", "true");
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
