@@ -8,7 +8,7 @@
 //  SS  SS   tt   ee      ll   ll  aa  aa
 //   SSSS     ttt  eeeee llll llll  aaaaa
 //
-// Copyright (c) 1995-2005 by Bradford W. Mott and the Stella team
+// Copyright (c) 1995-2006 by Bradford W. Mott and the Stella team
 //
 // See the file "license" for information on usage and redistribution of
 // this file, and for a DISCLAIMER OF ALL WARRANTIES.
@@ -26,7 +26,7 @@
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 SettingsWinCE::SettingsWinCE(OSystem* osystem) : Settings(osystem) 
 {
-  set("romdir", (string) getcwd() + '\\');
+  setInternal("romdir", (string) getcwd() + "\\Roms\\");
   setInt("wce_smartphone_paddlespeed", 3000);
 }
 
